# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for AQT configs."""
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import config
from aqt.jax.v2 import utils
import jax.numpy as jnp


def _dot_general_full_init_calibration(cfg):
  cfg.fwd.dg_quantizer.lhs.init_calibration()
  cfg.fwd.dg_quantizer.rhs.init_calibration()
  cfg.dlhs.dg_quantizer.lhs.init_calibration()
  cfg.dlhs.dg_quantizer.rhs.init_calibration()
  cfg.drhs.dg_quantizer.lhs.init_calibration()
  cfg.drhs.dg_quantizer.rhs.init_calibration()

# TODO(lew): We should use go/minion, go/blessing or other go/golden-diff-test


class AqtConfigTest(parameterized.TestCase):

  def _retrieve_quantizers(self, dot_general_raws):
    ret = []
    for dot_general_raw in dot_general_raws:
      ret.extend(
          [dot_general_raw.dg_quantizer.lhs, dot_general_raw.dg_quantizer.rhs]
      )
    return ret

  def test_config_v4(self):
    cfg = config.config_v4(
        fwd_bits=8,
        dlhs_bits=7,
        drhs_bits=6,
        rng_type='custom-1',
        dlhs_local_aqt=config.LocalAqt(contraction_axis_shard_count=2),
        drhs_local_aqt=config.LocalAqt(contraction_axis_shard_count=3),
        fwd_accumulator_dtype=jnp.int16,
        dlhs_accumulator_dtype=jnp.int8,
        drhs_accumulator_dtype=jnp.int4,
    )
    expected_cfg_str = """DotGeneral(fwd=DotGeneralRaw(lhs=Tensor(use_fwd_quant=False,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             rhs=Tensor(use_fwd_quant=False,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             dg_quantizer=DefaultDotGeneralQuantizer(lhs=Quantizer(numerics=IntSymmetric(bits=8,
                                                                                                         preserve_zero=True,
                                                                                                         preserve_max_val=False,
                                                                                                         clip=True,
                                                                                                         clip_gradient=False,
                                                                                                         round=True,
                                                                                                         noise_fn=None,
                                                                                                         dtype=<class 'jax.numpy.int8'>),
                                                                                   calib_shared_axes=None,
                                                                                   scale_stop_grad=True,
                                                                                   scale_dtype=None,
                                                                                   calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                   _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                 po2_scale=False,
                                                                                                                 clipping_scale=None),
                                                                                   context=Context(key=None,
                                                                                                   train_step=None,
                                                                                                   quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     rhs=Quantizer(numerics=IntSymmetric(bits=8,
                                                                                                         preserve_zero=True,
                                                                                                         preserve_max_val=False,
                                                                                                         clip=True,
                                                                                                         clip_gradient=False,
                                                                                                         round=True,
                                                                                                         noise_fn=None,
                                                                                                         dtype=<class 'jax.numpy.int8'>),
                                                                                   calib_shared_axes=None,
                                                                                   scale_stop_grad=True,
                                                                                   scale_dtype=None,
                                                                                   calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                   _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                 po2_scale=False,
                                                                                                                 clipping_scale=None),
                                                                                   context=Context(key=None,
                                                                                                   train_step=None,
                                                                                                   quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     lhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                           dtype=None),
                                                                                       calib_shared_axes=None,
                                                                                       scale_stop_grad=True,
                                                                                       scale_dtype=None,
                                                                                       calibration=None,
                                                                                       _calibrator=None,
                                                                                       context=Context(key=None,
                                                                                                       train_step=None,
                                                                                                       quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     rhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                           dtype=None),
                                                                                       calib_shared_axes=None,
                                                                                       scale_stop_grad=True,
                                                                                       scale_dtype=None,
                                                                                       calibration=None,
                                                                                       _calibrator=None,
                                                                                       context=Context(key=None,
                                                                                                       train_step=None,
                                                                                                       quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     lhs_mid_alpha=None,
                                                                     rhs_mid_alpha=None,
                                                                     skip_mid_scales=True),
                             dg_accumulator_dtype=<class 'jax.numpy.int16'>,
                             local_aqt=None,
                             jax_scope_name='aqt_fwd',
                             allow_dummy_gradient_into_qtensor=False,
                             dot_general=<function dot_general>),
           dlhs=DotGeneralRaw(lhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_quantizer=DefaultDotGeneralQuantizer(lhs=Quantizer(numerics=IntSymmetric(bits=7,
                                                                                                          preserve_zero=True,
                                                                                                          preserve_max_val=False,
                                                                                                          clip=True,
                                                                                                          clip_gradient=False,
                                                                                                          round=True,
                                                                                                          noise_fn=RandomCenteredUniform(),
                                                                                                          dtype=<class 'jax.numpy.int8'>),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                    _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                  po2_scale=False,
                                                                                                                  clipping_scale=None),
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs=Quantizer(numerics=IntSymmetric(bits=7,
                                                                                                          preserve_zero=True,
                                                                                                          preserve_max_val=False,
                                                                                                          clip=True,
                                                                                                          clip_gradient=False,
                                                                                                          round=True,
                                                                                                          noise_fn=None,
                                                                                                          dtype=<class 'jax.numpy.int8'>),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                    _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                  po2_scale=False,
                                                                                                                  clipping_scale=None),
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid_alpha=None,
                                                                      rhs_mid_alpha=None,
                                                                      skip_mid_scales=True),
                              dg_accumulator_dtype=<class 'jax.numpy.int8'>,
                              local_aqt=LocalAqt(contraction_axis_shard_count=2,
                                                 contraction_axis_shard_size=None,
                                                 tile_largest_shape=False),
                              jax_scope_name='aqt_dlhs',
                              allow_dummy_gradient_into_qtensor=False,
                              dot_general=<function dot_general>),
           drhs=DotGeneralRaw(lhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_quantizer=DefaultDotGeneralQuantizer(lhs=Quantizer(numerics=IntSymmetric(bits=6,
                                                                                                          preserve_zero=True,
                                                                                                          preserve_max_val=False,
                                                                                                          clip=True,
                                                                                                          clip_gradient=False,
                                                                                                          round=True,
                                                                                                          noise_fn=RandomCenteredUniform(),
                                                                                                          dtype=<class 'jax.numpy.int8'>),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                    _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                  po2_scale=False,
                                                                                                                  clipping_scale=None),
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs=Quantizer(numerics=IntSymmetric(bits=6,
                                                                                                          preserve_zero=True,
                                                                                                          preserve_max_val=False,
                                                                                                          clip=True,
                                                                                                          clip_gradient=False,
                                                                                                          round=True,
                                                                                                          noise_fn=None,
                                                                                                          dtype=<class 'jax.numpy.int8'>),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                    _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                  po2_scale=False,
                                                                                                                  clipping_scale=None),
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid_alpha=None,
                                                                      rhs_mid_alpha=None,
                                                                      skip_mid_scales=True),
                              dg_accumulator_dtype=<class 'jax.numpy.int4'>,
                              local_aqt=LocalAqt(contraction_axis_shard_count=3,
                                                 contraction_axis_shard_size=None,
                                                 tile_largest_shape=False),
                              jax_scope_name='aqt_drhs',
                              allow_dummy_gradient_into_qtensor=False,
                              dot_general=<function dot_general>),
           apply_custom_vjp_on_jax=True)"""
    _dot_general_full_init_calibration(cfg)
    utils.test_pprint_eq(cfg, expected_cfg_str, remove_memory_addresses=True)

  def test_config_v4_original(self):
    expected_cfg_str = """DotGeneral(fwd=DotGeneralRaw(lhs=Tensor(use_fwd_quant=False,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             rhs=Tensor(use_fwd_quant=False,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             dg_quantizer=DefaultDotGeneralQuantizer(lhs=Quantizer(numerics=IntSymmetric(bits=8,
                                                                                                         preserve_zero=True,
                                                                                                         preserve_max_val=False,
                                                                                                         clip=True,
                                                                                                         clip_gradient=False,
                                                                                                         round=True,
                                                                                                         noise_fn=None,
                                                                                                         dtype=<class 'jax.numpy.int8'>),
                                                                                   calib_shared_axes=None,
                                                                                   scale_stop_grad=True,
                                                                                   scale_dtype=None,
                                                                                   calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                   _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                 po2_scale=False,
                                                                                                                 clipping_scale=None),
                                                                                   context=Context(key=None,
                                                                                                   train_step=None,
                                                                                                   quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     rhs=Quantizer(numerics=IntSymmetric(bits=8,
                                                                                                         preserve_zero=True,
                                                                                                         preserve_max_val=False,
                                                                                                         clip=True,
                                                                                                         clip_gradient=False,
                                                                                                         round=True,
                                                                                                         noise_fn=None,
                                                                                                         dtype=<class 'jax.numpy.int8'>),
                                                                                   calib_shared_axes=None,
                                                                                   scale_stop_grad=True,
                                                                                   scale_dtype=None,
                                                                                   calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                   _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                 po2_scale=False,
                                                                                                                 clipping_scale=None),
                                                                                   context=Context(key=None,
                                                                                                   train_step=None,
                                                                                                   quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     lhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                           dtype=None),
                                                                                       calib_shared_axes=None,
                                                                                       scale_stop_grad=True,
                                                                                       scale_dtype=None,
                                                                                       calibration=None,
                                                                                       _calibrator=None,
                                                                                       context=Context(key=None,
                                                                                                       train_step=None,
                                                                                                       quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     rhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                           dtype=None),
                                                                                       calib_shared_axes=None,
                                                                                       scale_stop_grad=True,
                                                                                       scale_dtype=None,
                                                                                       calibration=None,
                                                                                       _calibrator=None,
                                                                                       context=Context(key=None,
                                                                                                       train_step=None,
                                                                                                       quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     lhs_mid_alpha=None,
                                                                     rhs_mid_alpha=None,
                                                                     skip_mid_scales=True),
                             dg_accumulator_dtype=<class 'jax.numpy.int32'>,
                             local_aqt=None,
                             jax_scope_name='aqt_fwd',
                             allow_dummy_gradient_into_qtensor=False,
                             dot_general=<function dot_general>),
           dlhs=DotGeneralRaw(lhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_quantizer=DefaultDotGeneralQuantizer(lhs=Quantizer(numerics=IntSymmetric(bits=8,
                                                                                                          preserve_zero=True,
                                                                                                          preserve_max_val=False,
                                                                                                          clip=True,
                                                                                                          clip_gradient=False,
                                                                                                          round=True,
                                                                                                          noise_fn=JaxUniform(),
                                                                                                          dtype=<class 'jax.numpy.int8'>),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                    _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                  po2_scale=False,
                                                                                                                  clipping_scale=None),
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs=Quantizer(numerics=IntSymmetric(bits=8,
                                                                                                          preserve_zero=True,
                                                                                                          preserve_max_val=False,
                                                                                                          clip=True,
                                                                                                          clip_gradient=False,
                                                                                                          round=True,
                                                                                                          noise_fn=None,
                                                                                                          dtype=<class 'jax.numpy.int8'>),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                    _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                  po2_scale=False,
                                                                                                                  clipping_scale=None),
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid_alpha=None,
                                                                      rhs_mid_alpha=None,
                                                                      skip_mid_scales=True),
                              dg_accumulator_dtype=<class 'jax.numpy.int32'>,
                              local_aqt=None,
                              jax_scope_name='aqt_dlhs',
                              allow_dummy_gradient_into_qtensor=False,
                              dot_general=<function dot_general>),
           drhs=DotGeneralRaw(lhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_quantizer=DefaultDotGeneralQuantizer(lhs=Quantizer(numerics=NoNumerics(noise_fn=JaxUniform(),
                                                                                                        dtype=None),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=None,
                                                                                    _calibrator=None,
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                        dtype=None),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=None,
                                                                                    _calibrator=None,
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid_alpha=None,
                                                                      rhs_mid_alpha=None,
                                                                      skip_mid_scales=True),
                              dg_accumulator_dtype=None,
                              local_aqt=None,
                              jax_scope_name='aqt_drhs',
                              allow_dummy_gradient_into_qtensor=False,
                              dot_general=<function dot_general>),
           apply_custom_vjp_on_jax=True)"""
    cfg = config.config_v4()
    _dot_general_full_init_calibration(cfg)
    utils.test_pprint_eq(cfg, expected_cfg_str, remove_memory_addresses=True)

  def test_config_fwd_fp8(self):
    expected_cfg = """DotGeneral(fwd=DotGeneralRaw(lhs=Tensor(use_fwd_quant=False,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             rhs=Tensor(use_fwd_quant=False,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             dg_quantizer=DefaultDotGeneralQuantizer(lhs=Quantizer(numerics=Fp8Numerics(dtype=<class 'jax.numpy.float8_e4m3fn'>,
                                                                                                        exponent_bits=4,
                                                                                                        mantissa_bits=3,
                                                                                                        stochastic_rounding=False),
                                                                                   calib_shared_axes=None,
                                                                                   scale_stop_grad=True,
                                                                                   scale_dtype=None,
                                                                                   calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                   _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                 po2_scale=False,
                                                                                                                 clipping_scale=None),
                                                                                   context=Context(key=None,
                                                                                                   train_step=None,
                                                                                                   quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     rhs=Quantizer(numerics=Fp8Numerics(dtype=<class 'jax.numpy.float8_e4m3fn'>,
                                                                                                        exponent_bits=4,
                                                                                                        mantissa_bits=3,
                                                                                                        stochastic_rounding=False),
                                                                                   calib_shared_axes=None,
                                                                                   scale_stop_grad=True,
                                                                                   scale_dtype=None,
                                                                                   calibration=<class 'aqt.jax.v2.calibration.AbsMaxCalibration'>,
                                                                                   _calibrator=AbsMaxCalibration(dtype=None,
                                                                                                                 po2_scale=False,
                                                                                                                 clipping_scale=None),
                                                                                   context=Context(key=None,
                                                                                                   train_step=None,
                                                                                                   quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     lhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                           dtype=None),
                                                                                       calib_shared_axes=None,
                                                                                       scale_stop_grad=True,
                                                                                       scale_dtype=None,
                                                                                       calibration=None,
                                                                                       _calibrator=None,
                                                                                       context=Context(key=None,
                                                                                                       train_step=None,
                                                                                                       quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     rhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                           dtype=None),
                                                                                       calib_shared_axes=None,
                                                                                       scale_stop_grad=True,
                                                                                       scale_dtype=None,
                                                                                       calibration=None,
                                                                                       _calibrator=None,
                                                                                       context=Context(key=None,
                                                                                                       train_step=None,
                                                                                                       quant_mode=<QuantMode.TRAIN: 1>)),
                                                                     lhs_mid_alpha=None,
                                                                     rhs_mid_alpha=None,
                                                                     skip_mid_scales=True),
                             dg_accumulator_dtype=<class 'jax.numpy.float32'>,
                             local_aqt=None,
                             jax_scope_name='aqt_fwd',
                             allow_dummy_gradient_into_qtensor=False,
                             dot_general=<function dot_general>),
           dlhs=DotGeneralRaw(lhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_quantizer=DefaultDotGeneralQuantizer(lhs=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                        dtype=None),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=None,
                                                                                    _calibrator=None,
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                        dtype=None),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=None,
                                                                                    _calibrator=None,
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid_alpha=None,
                                                                      rhs_mid_alpha=None,
                                                                      skip_mid_scales=True),
                              dg_accumulator_dtype=None,
                              local_aqt=None,
                              jax_scope_name='aqt_dlhs',
                              allow_dummy_gradient_into_qtensor=False,
                              dot_general=<function dot_general>),
           drhs=DotGeneralRaw(lhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_quantizer=DefaultDotGeneralQuantizer(lhs=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                        dtype=None),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=None,
                                                                                    _calibrator=None,
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                        dtype=None),
                                                                                    calib_shared_axes=None,
                                                                                    scale_stop_grad=True,
                                                                                    scale_dtype=None,
                                                                                    calibration=None,
                                                                                    _calibrator=None,
                                                                                    context=Context(key=None,
                                                                                                    train_step=None,
                                                                                                    quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      rhs_mid=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                                            dtype=None),
                                                                                        calib_shared_axes=None,
                                                                                        scale_stop_grad=True,
                                                                                        scale_dtype=None,
                                                                                        calibration=None,
                                                                                        _calibrator=None,
                                                                                        context=Context(key=None,
                                                                                                        train_step=None,
                                                                                                        quant_mode=<QuantMode.TRAIN: 1>)),
                                                                      lhs_mid_alpha=None,
                                                                      rhs_mid_alpha=None,
                                                                      skip_mid_scales=True),
                              dg_accumulator_dtype=None,
                              local_aqt=None,
                              jax_scope_name='aqt_drhs',
                              allow_dummy_gradient_into_qtensor=False,
                              dot_general=<function dot_general>),
           apply_custom_vjp_on_jax=True)"""
    cfg = config.config_fwd_fp8()
    _dot_general_full_init_calibration(cfg)
    utils.test_pprint_eq(cfg, expected_cfg, remove_memory_addresses=True)

  def test_set_int_numerics_preserve_zero(self):
    cfg = config.config_v4(drhs_bits=8)
    for quantizer in self._retrieve_quantizers([cfg.fwd, cfg.dlhs]):
      self.assertTrue(quantizer.numerics.preserve_zero)
      self.assertEqual(quantizer.numerics.dtype, jnp.int8)

    config.set_int_numerics_preserve_zero(cfg, preserve_zero=False)
    for quantizer in self._retrieve_quantizers([cfg.fwd, cfg.dlhs]):
      self.assertFalse(quantizer.numerics.preserve_zero)
      self.assertIsNone(quantizer.numerics.dtype)

  def test_set_absmax_calib_scale(self):
    cfg = config.config_v4(drhs_bits=8)
    for quantizer in self._retrieve_quantizers([cfg.fwd, cfg.dlhs, cfg.drhs]):
      self.assertIsNone(quantizer.calibration().clipping_scale)

    for quantizer in self._retrieve_quantizers([cfg.fwd, cfg.dlhs]):
      self.assertFalse(quantizer.numerics.clip_gradient)

    config.set_absmax_calib_scale(cfg, scale=3)
    for quantizer in self._retrieve_quantizers([cfg.fwd, cfg.dlhs, cfg.drhs]):
      self.assertAlmostEqual(quantizer.calibration().clipping_scale, 3)

    for quantizer in self._retrieve_quantizers([cfg.fwd, cfg.dlhs]):
      self.assertFalse(quantizer.numerics.clip_gradient)

    config.set_absmax_calib_scale(cfg, scale=0.1)
    for quantizer in self._retrieve_quantizers([cfg.fwd, cfg.dlhs]):
      self.assertTrue(quantizer.numerics.clip_gradient)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machine Translation example.

This script trains a Transformer on a WMT dataset.
"""
import concurrent.futures
import dataclasses
import functools
import json
import os
import os.path
import pathlib
import pickle
import re
import time
from typing import Any, Dict, Iterable, List, Mapping, Optional, Type, TypeVar, Union

from absl import app
from absl import flags
from absl import logging

from aqt.jax_legacy.jax import compute_cost_utils
from aqt.jax_legacy.jax import hlo_utils
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import train_utils


from aqt.jax_legacy.jax.wmt_mlperf import bleu
from aqt.jax_legacy.jax.wmt_mlperf import input_pipeline
from aqt.jax_legacy.jax.wmt_mlperf import models
from aqt.jax_legacy.jax.wmt_mlperf import predict

# train_flags is imported for the side effect of flag definitions.
# These are used both in xm_launch.py and this training executable.
from aqt.jax_legacy.jax.wmt_mlperf import train_flags  # pylint: disable=unused-import
from aqt.jax_legacy.jax.wmt_mlperf import training_hparams
from aqt.jax_legacy.utils import hparams_utils
from aqt.jax_legacy.utils import report_utils
from aqt.jax_legacy.utils import summary_utils

import flax
from flax import jax_utils
from flax import optim
from flax.metrics import tensorboard
from flax.training import checkpoints
from flax.training import common_utils
import jax
from jax import random
import jax.nn
import jax.numpy as jnp
from ml_collections import config_flags
import numpy as np
import sacrebleu
import tensorflow.compat.v2 as tf
import tensorflow_text as tftxt



flax.config.update('flax_use_orbax_checkpointing', False)

T = TypeVar('T')

EOS_TOKEN = 2  # Default Sentencepiece EOS token.

COMPUTE_MEMORY_COST_FILENAME = 'compute_memory_cost.json'

FLAGS = flags.FLAGS


flags.DEFINE_bool(
    'log_sparsity_scalars',
    default=True,
    help=(
        'Log `apply_sparsity`, `update_weight_sparsity` and '
        '`update_act_sparsity`  scalars to Tensorboard'
    ),
)

config_flags.DEFINE_config_file('hparams_config_dict', None,
                                'Path to file defining a config dict.')

flags.DEFINE_integer(
    'config_idx',
    default=None,
    help=(
        'Identifies which config within the sweep this training run should use.'
    ))

flags.DEFINE_string(
    'model_dir', default=None, help='Directory to store model data')

# More flags defined in train_flags.py.


def create_learning_rate_scheduler(
    hparams: training_hparams.LearningRateSchedulerHParams):
  """creates learning rate schedule.

  Args:
    hparams: A LearningRateSchedulerHParams instance that describes the desired
      learning rate schedule. See LearningRateSchedulerHParams documentation for
      details.

  Returns:
    a function learning_rate(step): float -> {'learning_rate': float}, the
    step-dependent lr.
  """
  factors = [n.strip() for n in hparams.factors.split('*')]

  def step_fn(step):
    """Step to learning rate function."""
    ret = 1.0
    for name in factors:
      if name == 'constant':
        ret *= hparams.base_learning_rate
      elif name == 'linear_warmup':
        ret *= jnp.minimum(1.0, step / hparams.warmup_steps)
      elif name == 'rsqrt_decay':
        ret /= jnp.sqrt(jnp.maximum(step, hparams.warmup_steps))
      elif name == 'rsqrt_normalized_decay':
        ret *= jnp.sqrt(hparams.warmup_steps)
        ret /= jnp.sqrt(jnp.maximum(step, hparams.warmup_steps))
      elif name == 'decay_every':
        # TODO(malmaud): Create an assert-style function to replace all these
        # if cond:
        #   raise ValueError
        # patterns.
        if hparams.decay_factor is None or hparams.steps_per_decay is None:
          raise ValueError('Missing parameters for "decay_every" factor.')
        ret *= (hparams.decay_factor**(step // hparams.steps_per_decay))
      elif name == 'cosine_decay':
        if hparams.steps_per_cycle is None:
          raise ValueError('Missing parameters for "cosine_decay" factor.')
        progress = jnp.maximum(0.0, (step - hparams.warmup_steps) /
                               float(hparams.steps_per_cycle))
        ret *= jnp.maximum(0.0,
                           0.5 * (1.0 + jnp.cos(jnp.pi * (progress % 1.0))))
      else:
        raise ValueError('Unknown factor %s.' % name)
    return jnp.asarray(ret, dtype=jnp.float32)

  return step_fn


@functools.partial(jax.jit, static_argnums=(1, 2, 3, 4))
def create_model(key, input_shape, target_shape,
                 transformer_kwargs: Mapping[str, Any],
                 hparams: models.Transformer.HParams):
  """Instantiate transformer model."""
  model = models.Transformer(
      **transformer_kwargs,
      dynamic_context=quant_config.DynamicContext(
          update_bounds=False, collect_acts_stats=FLAGS.collect_acts_stats),
      hparams=hparams,
      use_bfloat16=False,
      train=False,
      dropout_rate=0.1,
      attention_dropout_rate=0.1,
      should_decode=False)
  variables = model.init(
      key,
      jnp.zeros(input_shape, jnp.float32),
      jnp.zeros(target_shape, jnp.float32),
  )
  init_state, params = flax.core.pop(variables, 'params')  # pytype: disable=attribute-error
  return params, init_state


# TODO(malmaud): Consider bundling the optimizer parameters into their own
# HParams class
def create_optimizer(params, learning_rate, weight_decay, beta1, beta2, eps):
  optimizer_def = optim.Adam(
      learning_rate,
      beta1=beta1,
      beta2=beta2,
      eps=eps,
      weight_decay=weight_decay)
  optimizer = optimizer_def.create(params)
  optimizer = optimizer.replicate()
  return optimizer


@jax.custom_gradient
def cross_entropy_with_logits(logits, targets):
  """Computes cross entropy loss with custom gradient support.

  Args:
    logits: [batch * length, num_classes] float array.
    targets: categorical targets [batch * length] float array.

  Returns:
    Tuple of scalar loss and custom gradient function.
  """
  shifted = logits - logits.max(axis=-1, keepdims=True)
  exp_shifted = jnp.exp(shifted)
  sum_exp = jnp.sum(exp_shifted, axis=-1, keepdims=True)
  log_softmax = shifted - jnp.log(sum_exp)
  loss = - targets * log_softmax

  def grad_fn(g):
    return g * (exp_shifted / sum_exp - targets), g

  return loss, grad_fn


def compute_weighted_cross_entropy(logits,
                                   targets,
                                   weights=None,
                                   label_smoothing=0.1):
  """Compute weighted cross entropy and entropy for log probs and targets.

  Args:
    logits: [batch * length, num_classes] float array.
    targets: categorical targets [batch, length] int array.
    weights: None or array of shape [batch, length]
    label_smoothing: label smoothing constant, used to determine the on and off
      values.

  Returns:
    Tuple of scalar loss and batch normalizing factor.
  """
  targets = targets.reshape((-1))
  weights = weights.reshape((-1))
  if logits.ndim != targets.ndim + 1:
    raise ValueError('Incorrect shapes. Got shape %s logits and %s targets' %
                     (str(logits.shape), str(targets.shape)))
  vocab_size = logits.shape[-1]
  confidence = 1.0 - label_smoothing
  low_confidence = (1.0 - confidence) / (vocab_size - 1)
  normalizing_constant = -(
      confidence * jnp.log(confidence) +
      (vocab_size - 1) * low_confidence * jnp.log(low_confidence + 1e-20))
  soft_targets = common_utils.onehot(
      targets, vocab_size, on_value=confidence, off_value=low_confidence)

  loss = jnp.sum(cross_entropy_with_logits(logits, soft_targets), axis=-1)

  loss = loss - normalizing_constant

  normalizing_factor = np.prod(targets.shape)
  if weights is not None:
    loss = loss * weights
    normalizing_factor = weights.sum()

  return loss.sum(), normalizing_factor


def compute_weighted_accuracy(logits, targets, weights=None):
  """Compute weighted accuracy for log probs and targets.

  Args:
   logits: [batch * length, num_classes] float array.
   targets: categorical targets [batch, length] int array.
   weights: None or array of shape [batch, length]

  Returns:
    Tuple of scalar loss and batch normalizing factor.
  """
  targets = targets.reshape((-1))
  weights = weights.reshape((-1))
  if logits.ndim != targets.ndim + 1:
    raise ValueError('Incorrect shapes. Got shape %s logits and %s targets' %
                     (str(logits.shape), str(targets.shape)))
  loss = jnp.equal(jnp.argmax(logits, axis=-1), targets)
  normalizing_factor = np.prod(logits.shape[:-1])
  if weights is not None:
    loss = loss * weights
    normalizing_factor = weights.sum()

  return loss.sum(), normalizing_factor


def compute_metrics(logits, labels, weights):
  """Compute summary metrics."""
  loss, weight_sum = compute_weighted_cross_entropy(logits, labels, weights)
  acc, _ = compute_weighted_accuracy(logits, labels, weights)
  metrics = {
      'loss': loss,
      'accuracy': acc,
      'denominator': weight_sum,
  }
  metrics = jax.lax.psum(metrics, axis_name='batch')
  return metrics


# TODO(shivaniagrawal): Make TrainingHparams hashable instead.
class WrapHashably:
  """Wrapper to turn np.ndarray into Hashable objects for pmap."""
  __slots__ = ['val']

  def __init__(self, val):
    self.val = val

  def __hash__(self):
    return id(self.val)

  def __eq__(self, other):
    return self.val is other.val


def train_step(optimizer,
               batch,
               dynamic_context: quant_config.DynamicContext,
               transformer_kwargs: Mapping[str, Any],
               hparams: Union[training_hparams.TrainingHParams, WrapHashably],
               state,
               dropout_rng=None):
  """Perform a single training step."""
  if isinstance(hparams, WrapHashably):
    hparams = hparams.val

  train_keys = [
      'inputs', 'targets', 'inputs_position', 'targets_position',
      'inputs_segmentation', 'targets_segmentation'
  ]
  (inputs, targets, inputs_positions, targets_positions, inputs_segmentation,
   targets_segmentation) = [batch.get(k, None) for k in train_keys]

  weights = jnp.where(targets > 0, 1, 0).astype(jnp.float32)

  # It's very important to handle PRNG splitting inside the top pmap, rather
  # than handling it outside in the training loop - doing the latter can add
  # bad stalls to the input data transfer.
  dropout_rng, new_dropout_rng = random.split(dropout_rng)

  def loss_fn(params):
    """loss function used for training."""
    model = models.Transformer(
        **transformer_kwargs,
        dynamic_context=dynamic_context,
        use_bfloat16=FLAGS.use_bfloat16,
        train=True,
        hparams=hparams.model_hparams,
        dropout_rate=0.1,
        attention_dropout_rate=0.1,
        should_decode=False)
    logits, new_variables = model.apply(
        {
            'params': params,
            **state
        },
        inputs,
        targets,
        inputs_positions=inputs_positions,
        targets_positions=targets_positions,
        inputs_segmentation=inputs_segmentation,
        targets_segmentation=targets_segmentation,
        mutable=True,
        rngs={'dropout': dropout_rng})
    new_state, _ = flax.core.pop(new_variables, 'params')
    loss, weight_sum = compute_weighted_cross_entropy(logits, targets, weights)
    mean_loss = loss / weight_sum
    total_loss = mean_loss


    return total_loss, (logits, new_state)

  step = optimizer.state.step
  learning_rate_fn = create_learning_rate_scheduler(
      hparams.learning_rate_schedule)
  lr = learning_rate_fn(step)
  value_and_grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
  (_, (logits, new_state)), loss_grad = value_and_grad_fn(optimizer.target)
  new_optimizer = optimizer.apply_gradient(loss_grad, learning_rate=lr)

  if FLAGS.compute_train_metrics:
    metrics = compute_metrics(logits, targets, weights)
  else:
    metrics = {}
  metrics['learning_rate'] = lr
  metrics['apply_sparsity'] = dynamic_context.apply_sparsity
  metrics['num_update_sparsity'] = dynamic_context.num_update_sparsity
  metrics['update_weight_sparsity'] = dynamic_context.update_weight_sparsity
  metrics['update_act_sparsity'] = dynamic_context.update_act_sparsity


  return new_state, new_optimizer, metrics, new_dropout_rng

# TODO(shivaniagrawal): parametrize the axis_name and use the same axis name
# throughout Transformer model.
p_train_step = jax.pmap(
    functools.partial(train_step),
    axis_name='batch',
    static_broadcasted_argnums=(3, 4))
# hparams, transformer_kwargs are static args.


def _write_beam_hlo(params, input_shape, cache, state,
                    transformer_kwargs: Mapping[str, Any],
                    hparams: models.Transformer.HParams,
                    dynamic_context: quant_config.DynamicContext):
  """Writes HLO with beam search for given model."""
  if not FLAGS.output_beam_hlo_filename:
    return
  basename, _ = os.path.splitext(FLAGS.output_beam_hlo_filename)
  input_dummy = jnp.ones(input_shape)

  def _without_weights(inputs, params):
    """JAX computation with inputs and weights turned into parameters."""
    return predict_step(
        inputs,
        params,
        cache,
        state,
        EOS_TOKEN,
        transformer_kwargs=transformer_kwargs,
        hparams=hparams,
        dynamic_context=dynamic_context)

  def _with_weights(inputs):
    """JAX computation with only inputs turned into parameters."""
    return predict_step(
        inputs,
        params,
        cache,
        state,
        EOS_TOKEN,
        transformer_kwargs=transformer_kwargs,
        hparams=hparams,
        dynamic_context=dynamic_context)

  _write_hlo(basename, _without_weights, input_dummy, transformer_kwargs,
             params, cache, state)
  _write_hlo(basename + '_with_weights', _with_weights, input_dummy, state)


def get_jax_computation_of_model(transformer_kwargs: Mapping[str, Any],
                                 params: Mapping[str, Any], state: Mapping[str,
                                                                           Any],
                                 hparams: training_hparams.TrainingHParams,
                                 with_weights: bool):
  """Creates jax computation functions of model with or without weights."""
  input_dummy = jnp.ones((hparams.per_host_batch_size, FLAGS.max_target_length))
  target_dummy = jnp.ones(
      (hparams.per_host_batch_size, FLAGS.max_target_length))
  model = models.Transformer(
      **transformer_kwargs,
      dynamic_context=quant_config.DynamicContext(
          update_bounds=False, collect_acts_stats=False),
      train=False,
      hparams=hparams.model_hparams,
      use_bfloat16=False,
      dropout_rate=0.1,
      attention_dropout_rate=0.1,
      should_decode=False)

  def _without_weights(params, inputs, target, state):
    """JAX computation with inputs and weights turned into parameters."""
    logits = model.apply({
        'params': params,
        **state
    },
                         inputs,
                         target,
                         mutable=False)
    return logits

  def _with_weights(inputs, target, state):
    """JAX computation with only inputs turned into parameters."""
    logits = model.apply({'params': params, **state}, inputs, target)
    return logits

  if with_weights:
    return _with_weights, input_dummy, target_dummy, state, params
  else:
    return _without_weights, input_dummy, target_dummy, state, params


def estimate_compute_and_memory_cost(
    model_dir: str, params: Mapping[str, Any], transformer_kwargs: Mapping[str,
                                                                           Any],
    state: Mapping[str,
                   Any], hparams: training_hparams.TrainingHParams) -> None:
  """Estimate compute and memory cost of model."""

  without_weights_fn, input_dummy, target_dummy, state, params = (
      get_jax_computation_of_model(
          params=params,
          transformer_kwargs=transformer_kwargs,
          state=state,
          hparams=hparams,
          with_weights=False,
      )
  )
  cost_dict = compute_cost_utils.estimate_costs_of_dot_and_conv_ops_from_jax_fn(
      without_weights_fn, params, input_dummy, target_dummy, state)

  path = tf.io.gfile.join(model_dir, COMPUTE_MEMORY_COST_FILENAME)
  with tf.io.gfile.GFile(path, 'w') as file:
    json.dump(cost_dict, file, indent=2)
  logging.info('Estimated compute and memory costs and wrote to file')


def _write_train_hlo(transformer_kwargs: Mapping[str, Any],
                     params: Mapping[str, Any], state: Mapping[str, Any],
                     hparams: training_hparams.TrainingHParams):
  """Writes training HLO for given model."""
  if not FLAGS.output_hlo_filename:
    return

  basename, _ = os.path.splitext(FLAGS.output_hlo_filename)
  without_weights_fn, input_dummy, target_dummy, state, model = (
      get_jax_computation_of_model(
          params=params,
          transformer_kwargs=transformer_kwargs,
          state=state,
          hparams=hparams,
          with_weights=False,
      )
  )
  _write_hlo(basename, without_weights_fn, model, input_dummy, target_dummy,
             state)
  with_weights_fn, input_dummy, target_dummy, state, model = (
      get_jax_computation_of_model(
          params=params,
          transformer_kwargs=transformer_kwargs,
          state=state,
          hparams=hparams,
          with_weights=True,
      )
  )
  _write_hlo(basename + '_with_weights', with_weights_fn, input_dummy,
             target_dummy, state)


def _write_hlo(output, fn, *fn_args, **fn_kwargs):
  """Writes HLO for the given function.

  Writes the HLO of the provided function to destination
  FLAGS.model_dir/<output>{.txt,.pb}. Both the HLO-text and binary protobuf are
  written.

  Args:
    output: The basename of the output file (without extension)
    fn: the function for which the HLO is to be produced.
    *fn_args: the function's args.
    **fn_kwargs: the function's kwargs.

  Returns:
    None.
  """
  try:
    computation = (
        jax.jit(fn).lower(*fn_args, **fn_kwargs).compiler_ir(dialect='hlo')
    )
    logging.info('Generated XLA computation for HLO.')
    pb_path = tf.io.gfile.join(FLAGS.model_dir, f'{output}.pb')
    if not tf.io.gfile.exists(pb_path):
      hlo_utils.output_hlo(computation, pb_path)
      logging.info('Wrote serialized proto %s file.', pb_path)
    txt_path = tf.io.gfile.join(FLAGS.model_dir, f'{output}.txt')
    if not tf.io.gfile.exists(txt_path):
      hlo_utils.output_hlo(computation, txt_path)
      logging.info('Wrote HLO-text %s file.', txt_path)
  except Exception:  # pylint: disable=broad-except
    logging.exception('Exception in _write_hlo.')


def eval_step(params: Mapping[str, Any], batch, state: Mapping[str, Any],
              transformer_kwargs: Mapping[str, Any],
              hparams: models.Transformer.HParams,
              dynamic_context: quant_config.DynamicContext):
  """Calculate evaluation metrics on a batch."""
  inputs, targets = batch['inputs'], batch['targets']
  weights = jnp.where(targets > 0, 1.0, 0.0)
  model = models.Transformer(
      **transformer_kwargs,
      dynamic_context=dynamic_context,
      use_bfloat16=FLAGS.use_bfloat16,
      train=False,
      hparams=hparams,
      dropout_rate=0.0,
      attention_dropout_rate=0.0,
      should_decode=False)
  mutable = ['sparsity']
  # .apply() will return a tuple if mutable is not False
  logits, _ = model.apply({
      'params': params,
      **state
  },
                          inputs,
                          targets,
                          mutable=mutable)
  return compute_metrics(logits, targets, weights)


p_eval_step = jax.pmap(
    eval_step, axis_name='batch', static_broadcasted_argnums=(3, 4))
# Argument 3 is transformer_kwargs. Argument 4 is hparams.


def predict_step(inputs, params, cache, state, eos_token,
                 transformer_kwargs: Mapping[str, Any],
                 hparams: models.Transformer.HParams,
                 dynamic_context: quant_config.DynamicContext):
  return predict.step(
      inputs,
      params,
      cache,
      state,
      eos_token,
      FLAGS.max_predict_length,
      transformer_kwargs=transformer_kwargs,
      hparams=hparams,
      dynamic_context=dynamic_context)


p_pred_step = jax.pmap(
    predict_step, axis_name='batch', static_broadcasted_argnums=(4, 5, 6))
# argument 4: eos token is constant. argument 5: transformer_kwargs. argument 6:
# hparams.


def pad_examples(x, desired_batch_size):
  """Expand batch to desired size by repeating last slice."""
  batch_pad = desired_batch_size - x.shape[0]
  return np.concatenate([x, np.tile(x[-1], (batch_pad, 1))], axis=0)


def tohost(x):
  """Collect batches from all devices to host and flatten batch dimensions."""
  n_device, n_batch, *remaining_dims = x.shape
  return np.array(x).reshape((n_device * n_batch,) + tuple(remaining_dims))


def run_eval(*,
             ds: Iterable[Any],
             state: Dict[Any, Any],
             num_steps: int,
             params: Mapping[str, Any],
             summary_writer: Optional[Any] = None,
             step: Optional[int] = None,
             transformer_kwargs: Mapping[str, Any],
             hparams: training_hparams.TrainingHParams) -> Dict[str, Any]:
  """Compute evaluation metrics for a given dataset."""
  eval_metrics = []
  eval_iter = iter(ds)
  dynamic_context = get_dynamic_context(hparams, step, train=False)
  for _, eval_batch in zip(range(num_steps), eval_iter):
    eval_batch = jax.tree.map(lambda x: x._numpy(), eval_batch)  # pylint: disable=protected-access
    eval_batch = common_utils.shard(eval_batch)
    metrics = p_eval_step(params, eval_batch, state, transformer_kwargs,
                          hparams.model_hparams, dynamic_context)
    eval_metrics.append(metrics)
  eval_metrics = common_utils.get_metrics(eval_metrics)
  eval_metrics_sums = jax.tree.map(jnp.sum, eval_metrics)
  eval_denominator = eval_metrics_sums.pop('denominator')
  eval_summary = jax.tree.map(
      lambda x: x / eval_denominator,  # pylint: disable=cell-var-from-loop
      eval_metrics_sums)
  eval_summary['perplexity'] = jnp.clip(
      jnp.exp(eval_summary['loss']), a_max=1.0e4)
  if summary_writer is not None:
    if jax.host_id() == 0:
      for key, val in eval_summary.items():
        summary_writer.scalar(key, val, step)
      summary_writer.flush()
  return eval_summary


# TODO(shivaniagrawal): This is set up to accommodate mocking of the encoder in
# train_test.py, which seems like it should be unnecessary.
def decode_tokens(toks, encoder):
  eos_tokens = (toks == EOS_TOKEN)
  if np.any(eos_tokens):
    valid_toks = toks[:np.argmax(eos_tokens) + 1].astype(np.int32)
  else:
    valid_toks = toks[:].astype(np.int32)
  return encoder.detokenize(valid_toks).numpy().decode('utf-8')


@functools.partial(jax.jit, static_argnums=(0, 1, 2))
def initialize_cache(batch_size: int, transformer_kwargs: Dict[str, Any],
                     hparams: models.Transformer.HParams):
  """Initialize a decoding code suitable for `run_inference`."""
  max_decode_len = transformer_kwargs['max_len']
  inputs_shape = (batch_size, 1)
  target_shape = (batch_size, max_decode_len)
  initial_variables = models.Transformer(
      **transformer_kwargs,
      hparams=hparams,
      should_decode=True,
      train=False,
      dynamic_context=quant_config.DynamicContext(
          update_bounds=False, collect_acts_stats=False),
      dropout_rate=0.0,
      attention_dropout_rate=0.0,
      use_bfloat16=False).init(
          jax.random.PRNGKey(0), jnp.ones(inputs_shape, jnp.int32),
          jnp.ones(target_shape, jnp.int32))
  return initial_variables['cache']


def run_inference(*, ds, transformer_kwargs,
                  hparams: training_hparams.TrainingHParams, state, encoder,
                  params, write_beam_hlo, step: int):
  """Runs inference (ie, translates) over the given dataset."""

  predict_iter = iter(ds)
  sources, references, predictions = [], [], []
  dynamic_context = get_dynamic_context(hparams, step, train=False)
  for _, pred_batch in enumerate(predict_iter):
    pred_batch = jax.tree.map(lambda x: x._numpy(), pred_batch)  # pylint: disable=protected-access
    # Handle final odd-sized batch by padding instead of dropping it.
    cur_pred_batch_size = pred_batch['inputs'].shape[0]
    if cur_pred_batch_size != FLAGS.eval_batch_size:
      logging.info('Translation: uneven batch size %d.', cur_pred_batch_size)
      pred_batch = jax.tree.map(
          lambda x: pad_examples(x, FLAGS.eval_batch_size), pred_batch)
    pred_batch = common_utils.shard(pred_batch)
    per_device_batchsize = pred_batch['inputs'].shape[1]
    unreplicated_cache = initialize_cache(per_device_batchsize,
                                          flax.core.freeze(transformer_kwargs),
                                          hparams.model_hparams)

    if write_beam_hlo:
      logging.info('Writing inference HLO.')
      params = jax_utils.unreplicate(params)
      state = jax_utils.unreplicate(state)
      logging.info('Unreplicated optimizer and state.')
      _write_beam_hlo(
          params=params,
          input_shape=pred_batch['inputs'].shape[1:],
          cache=unreplicated_cache,
          state=state,
          transformer_kwargs=transformer_kwargs,
          hparams=hparams.model_hparams,
          dynamic_context=dynamic_context)
      logging.info('Finished writing inference HLO.')
      params = jax_utils.replicate(params)
      state = jax_utils.replicate(state)
      logging.info('Replicated optimizer.')

    cache = jax_utils.replicate(unreplicated_cache)
    predicted = p_pred_step(
        pred_batch['inputs'],
        params,
        cache,
        state,
        EOS_TOKEN,
        transformer_kwargs,
        hparams.model_hparams,
        dynamic_context=dynamic_context)
    predicted = tohost(predicted)
    inputs = tohost(pred_batch['inputs'])
    targets = tohost(pred_batch['targets'])
    # Iterate through non-padding examples of batch.
    for i, s in enumerate(predicted[:cur_pred_batch_size]):
      sources.append(decode_tokens(inputs[i], encoder))
      references.append(decode_tokens(targets[i], encoder))
      predictions.append(decode_tokens(s, encoder))
  return sources, references, predictions


# TODO(malmaud): Make these types more precise
# TODO(malmaud): Expand the usage of this class throughout the training loop
@dataclasses.dataclass
class TrainingState:
  """Represents the state of the training loop at any given step."""
  flax_state: Any
  optimizer: Any
  dropout_rngs: Any
  transformer_kwargs: Any

  @classmethod
  def initialize(
      cls: Type[T],
      encoder: tftxt.SentencepieceTokenizer,
      hparams: training_hparams.TrainingHParams,
      transformer_kwargs_overrides: Optional[Dict[Any, Any]] = None) -> T:
    """Initializes a training state in the form of a TrainingState instance."""
    random_seed = hparams.random_seed
    rng = random.PRNGKey(random_seed)
    rng, init_rng = random.split(rng)
    batch_size = hparams.per_host_batch_size
    max_target_length = FLAGS.max_target_length
    max_eval_target_length = FLAGS.max_eval_target_length
    max_length = max(max_target_length, max_eval_target_length)
    input_shape = (batch_size, max_target_length)
    target_shape = (batch_size, max_target_length)
    vocab_size = int(encoder.vocab_size())

    transformer_kwargs = {
        'vocab_size': vocab_size,
        'output_vocab_size': vocab_size,
        'max_len': max_length,
    }

    if transformer_kwargs_overrides is not None:
      transformer_kwargs.update(transformer_kwargs_overrides)

    transformer_kwargs = flax.core.freeze(transformer_kwargs)
    params, init_state = create_model(init_rng, tuple(input_shape),
                                      tuple(target_shape), transformer_kwargs,
                                      hparams.model_hparams)
    init_state = jax_utils.replicate(init_state)
    optimizer = create_optimizer(
        params,
        learning_rate=hparams.learning_rate_schedule.base_learning_rate,
        weight_decay=hparams.weight_decay,
        beta1=hparams.beta1,
        beta2=hparams.beta2,
        eps=hparams.eps)

    # We init the first set of dropout PRNG keys, but update it afterwards
    # inside the main pmap'd training update for performance.
    dropout_rngs = random.split(rng, jax.local_device_count())
    return cls(
        optimizer=optimizer,
        flax_state=init_state,
        dropout_rngs=dropout_rngs,
        transformer_kwargs=transformer_kwargs)

  def save_checkpoint(self,
                      *,
                      model_dir: str,
                      step: float,
                      prefix: str = 'checkpoint_') -> None:
    """Saves TrainingState checkpoint to model_dir."""
    unreplicated_flax_state = jax_utils.unreplicate(self.flax_state)
    checkpoints.save_checkpoint(
        model_dir, (self.optimizer, unreplicated_flax_state, self.dropout_rngs,
                    self.transformer_kwargs),
        step,
        prefix=prefix)  # pytype: disable=wrong-arg-types

  def restore_checkpoint(self,
                         *,
                         model_dir: str,
                         prefix: str = 'checkpoint_') -> 'TrainingState':
    """Restore TrainingState from checkpoint in model_dir."""
    unreplicated_flax_state = jax_utils.unreplicate(self.flax_state)
    target_state = (self.optimizer, unreplicated_flax_state, self.dropout_rngs,
                    self.transformer_kwargs)
    try:
      optimizer, flax_state, dropout_rngs, transformer_kwargs = (
          checkpoints.restore_checkpoint(model_dir, target_state, prefix=prefix)
      )
    except ValueError:
      optimizer, flax_state, dropout_rngs, transformer_kwargs = target_state
    flax_state = jax_utils.replicate(flax_state)
    return type(self)(
        flax_state=flax_state,
        optimizer=optimizer,
        dropout_rngs=dropout_rngs,
        transformer_kwargs=transformer_kwargs)


def get_dynamic_context(hparams: training_hparams.TrainingHParams, step: int,
                        train: bool) -> quant_config.DynamicContext:
  """Returns quantization context for a given training step.

  Args:
    hparams: A TrainingHParams instance.
    step: Current training step.
    train: Boolean of whether this contexet will be used for a training or eval
      step.

  Returns:
    A DynamicContext instance that has been replicated so it can be passed to
      a pmapped function.
  """
  if train:
    collect_acts_stats = FLAGS.collect_acts_stats
  else:
    collect_acts_stats = False
  dynamic_context = train_utils.get_dynamic_context_for_step(
      activation_bound_update_freq=hparams.activation_bound_update_freq,
      activation_bound_start_step=hparams.activation_bound_start_step,
      # TODO(shivaniagrawal): make a decision on passing this via hparams.
      sparsity_start_step=FLAGS.sparsity_start_step,
      sparsity_update_freq=FLAGS.sparsity_update_freq,
      step=step,
      collect_acts_stats=collect_acts_stats,
      prefer_int8_to_int32_dot=hparams.prefer_int8_to_int32_dot)
  if not train:
    dynamic_context = dataclasses.replace(dynamic_context, update_bounds=False)
    dynamic_context = dataclasses.replace(dynamic_context, apply_sparsity=True)
    dynamic_context = dataclasses.replace(
        dynamic_context, update_act_sparsity=True)
    dynamic_context = dataclasses.replace(
        dynamic_context, update_weight_sparsity=False)
  return jax_utils.replicate(dynamic_context)


def run_train_step(*, training_state: TrainingState, step: int, batch: Any,
                   hparams: training_hparams.TrainingHParams):
  """Run a single step of training."""
  dynamic_context = get_dynamic_context(hparams, step, train=True)
  # Shard data to devices and do a training step.
  batch = common_utils.shard(jax.tree.map(lambda x: x._numpy(), batch))  # pylint: disable=protected-access
  flax_state, optimizer, metrics, dropout_rngs = p_train_step(
      training_state.optimizer,
      batch,
      dynamic_context,
      training_state.transformer_kwargs,
      WrapHashably(hparams),
      state=training_state.flax_state,
      dropout_rng=training_state.dropout_rngs)
  new_training_state = TrainingState(
      flax_state=flax_state,
      optimizer=optimizer,
      dropout_rngs=dropout_rngs,
      transformer_kwargs=training_state.transformer_kwargs)
  return new_training_state, metrics


@dataclasses.dataclass
class Datasets:
  """Represents all the datasets used by the training loop.

  Datasets are assumed to share a vocabularity, which is encapsulated in the
  encoder attribute.
  """
  train_ds: tf.data.Dataset  # Training dataset
  # Evaluation dataset dictionary in the format {name: dataset} for calculating
  # perplexity.
  eval_ds_dict: Dict[str, tf.data.Dataset]
  train_eval_ds: tf.data.Dataset  # Sample of training dataset for evaluation
  # Prediction dataset dictionary in the format {name: dataset} for calculating
  # BLEU scores. Same as the eval dataset list, but preprocessed slightly
  # differently.
  predict_ds_dict: Dict[str, tf.data.Dataset]
  # Encoder for the vocabulary shared all the preceding datasets.
  encoder: tftxt.SentencepieceTokenizer

  @classmethod
  def load(cls: Type[T], random_seed: int,
           eval_dataset_list: Optional[List[str]], batch_size: int) -> T:
    """Loads all datasets used by the training loop."""
    n_devices = jax.local_device_count()
    vocab_path = FLAGS.vocab_path
    if vocab_path is None:
      vocab_path = os.path.join(FLAGS.model_dir, 'sentencepiece_model')
    train_ds, eval_ds_dict, train_eval_ds, predict_ds_dict, encoder = (
        input_pipeline.get_wmt_datasets(
            n_devices=n_devices,
            dataset_name=FLAGS.dataset_name,
            eval_dataset_list=eval_dataset_list,
            shard_idx=jax.host_id(),
            shard_count=jax.host_count(),
            data_dir=FLAGS.data_dir,
            vocab_path=vocab_path,
            batch_size=batch_size,
            max_length=FLAGS.max_target_length,
            max_eval_length=FLAGS.max_eval_target_length,
            seed=random_seed,
        )
    )
    return cls(
        train_ds=train_ds,
        eval_ds_dict=eval_ds_dict,
        train_eval_ds=train_eval_ds,
        predict_ds_dict=predict_ds_dict,
        encoder=encoder)




def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')
  # This seems to be necessary even when importing TF2?
  tf.enable_v2_behavior()
  if FLAGS.hparams_config_dict:
    # In this case, there are multiple training configs defined in the config
    # dict, so we pull out the one this training run should use.
    if 'configs' in FLAGS.hparams_config_dict:
      hparams_config_dict = FLAGS.hparams_config_dict.configs[FLAGS.config_idx]
    else:
      hparams_config_dict = FLAGS.hparams_config_dict
    hparams = hparams_utils.load_dataclass_from_config_dict(
        training_hparams.TrainingHParams, hparams_config_dict)


  # Number of local devices for this host.
  n_devices = jax.local_device_count()

  if hparams.per_host_batch_size % n_devices:
    raise ValueError(
        f'Batch size must be divisible by the number of devices. Got batch '
        f'size {hparams.per_host_batch_size} for {n_devices}.')

  logging.info('Initializing dataset.')
  eval_dataset_names_list = FLAGS.additional_eval_datasets
  if FLAGS.eval_dataset_name not in eval_dataset_names_list:
    eval_dataset_names_list.append(FLAGS.eval_dataset_name)

  datasets = Datasets.load(
      hparams.random_seed,
      eval_dataset_list=eval_dataset_names_list,
      batch_size=hparams.per_host_batch_size)

  with concurrent.futures.ThreadPoolExecutor(max_workers=1) as io_executor:
    run_training(datasets=datasets, hparams=hparams, io_executor=io_executor)



def save_best_checkpoint(*, model_dir: str, training_state: TrainingState,
                         loss: float) -> None:
  """Saves a checkpoint if 'loss' is smaller than the loss of the previous best-model checkpoint."""

  prefix = 'best_checkpoint_eval_loss_'

  # If the loss is NaN or inf, then this isn't the best checkpoint. We
  # explicitly return early since 'checkpoints.save_checkpoint' would overwrite
  # the best checkpoint with a bad checkpoint.
  if not jnp.isfinite(loss):
    return

  # Scan the existing checkpoints and see if any correspond to a lower loss. If
  # so, this isn't the best checkpoint so return. Otherwise, we have a new best
  # checkpoint so delete the current one and write a new one.

  for dir_name, _, file_names in tf.io.gfile.walk(model_dir):
    for filename in file_names:
      if filename == 'best_checkpoint_eval_loss_tmp':
        file_path = tf.io.gfile.join(dir_name, filename)
        tf.io.gfile.remove(file_path)
        return
      # This regex will match a filename like 'best_checkpoint_eval_loss_1.5'
      # and will extract '1.5' into 'current_best_loss'.
      checkpoint_match = re.match(rf'.*{prefix}(.*)$', str(filename))
      if checkpoint_match is not None:
        current_best_loss_str = checkpoint_match.group(1)
        if current_best_loss_str == 'tmp':
          return
        current_best_loss = float(current_best_loss_str)
        if loss >= current_best_loss:
          return
        else:
          file_path = tf.io.gfile.join(dir_name, filename)
          tf.io.gfile.remove(file_path)

  training_state.save_checkpoint(model_dir=model_dir, step=loss, prefix=prefix)


def get_state_dict_keys_from_flags():
  """Returns key suffixes to look up in flax state dict."""
  state_dict_keys = []
  if FLAGS.visualize_acts_bound:
    state_dict_keys.append('bounds')
  if FLAGS.collect_acts_stats:
    # State names from googlex.positron.tensorflow.jax.aqt.stats_tag.StatsTag
    state_dict_keys.extend([
        'min_per_ch', 'max_per_ch', 'mean_per_ch', 'stddev_per_ch',
        'absdev_per_ch', 'stddev_per_ch_uncentered', 'absdev_per_ch_uncentered',
    ])
  return state_dict_keys




def eval_ds_name_to_summary_dir(eval_dataset_name: str) -> str:
  return 'eval_' + str('--'.join(os.path.abspath(eval_dataset_name).parts))


def does_checkpoint_exist(model_dir: str) -> bool:
  """Determine if a checkpoint for this training run has already been saved."""
  # The model directory where checkpoints are saved doesn't exist yet.
  if not tf.io.gfile.exists(model_dir):
    return False
  for _, _, file_names in tf.io.gfile.walk(model_dir):
    for file in file_names:
      # This will match a checkpoint with a filename matching checkpoint_{step
      # number}, indicating a checkpoint already exists.
      if re.match(r'checkpoint_(\d+)', file):
        return True
  return False


def run_training(
    datasets: Datasets,
    hparams: training_hparams.TrainingHParams,
    io_executor: concurrent.futures.ThreadPoolExecutor,
    transformer_kwargs_overrides: Optional[Dict[str, Any]] = None,
):
  """Run the main training loop until completion."""
  hparams_utils.write_hparams_to_file_with_host_id_check(
      hparams, FLAGS.model_dir)

  # Use hardware RNG for bernoulli randoms in dropout mask creation.
  if hparams.hardware_rng:
    models.set_hardware_bernoulli()

  batch_size = hparams.per_host_batch_size
  num_train_steps = hparams.num_train_steps
  logging.info('num train steps: %d', num_train_steps)
  num_eval_steps = FLAGS.num_eval_steps
  eval_freq = FLAGS.eval_frequency
  stats_serialization_freq = FLAGS.stats_serialization_freq

  train_summary_writer = None
  eval_summary_writer = None
  additional_eval_summary_writers = {}
  train_eval_summary_writer = None
  if jax.host_id() == 0:
    model_dir = FLAGS.model_dir
    train_summary_writer = tensorboard.SummaryWriter(model_dir + '/train')
    eval_summary_writer = tensorboard.SummaryWriter(model_dir + '/eval')
    train_eval_summary_writer = tensorboard.SummaryWriter(model_dir +
                                                          '/train_eval')
    for eval_dataset_name in datasets.eval_ds_dict:
      if eval_dataset_name == FLAGS.eval_dataset_name:
        continue
      additional_eval_summary_writers[
          eval_dataset_name] = tensorboard.SummaryWriter(
              model_dir + '/' + eval_ds_name_to_summary_dir(eval_dataset_name))

  local_device_count = jax.local_device_count()
  if (batch_size % local_device_count or
      FLAGS.eval_batch_size % local_device_count):
    raise ValueError(
        f'Batch size must be divisible by the number of devices. Training '
        f'batch_size is {batch_size} and eval_batch_size is '
        f'{FLAGS.eval_batch_size} for number of devices {local_device_count}.')

  train_iter = iter(datasets.train_ds)
  start_step = 0
  training_state = TrainingState.initialize(
      encoder=datasets.encoder,
      hparams=hparams,
      transformer_kwargs_overrides=transformer_kwargs_overrides)

  # Load model state from a checkpoint. Summary of model restoration flowchart:
  # 1) If the training run is just starting up and so hasn't made any progress
  #    yet (as determined by the absence of a checkpoint in model_dir),
  #    we check if the user specified an initial checkpoint to start training
  #    from via the `restore_checkpoint_model_dir` flag.
  #
  #   1a) If so, we load that checkpoint.
  #
  #   1b) If not, we start the training from scratch.
  #
  # 2) If the training run was already in progress but was pre-empted by
  #    XManager and is now resuming (as determined by the presence of a saved
  #    checkpoint in the model directory), we resume training based on the
  #    latest saved checkpoint.
  #
  # In a typical usecase, 'restore_checkpoint_model_dir' might point to a
  # bfloat16 model directory trained to convergence and this training run will
  # perform quantization-aware finetuning of that model.
  checkpoint_model_dir = None
  if (
      FLAGS.restore_checkpoint_model_dir is not None
      and not does_checkpoint_exist(FLAGS.model_dir)
  ):
    checkpoint_model_dir = FLAGS.restore_checkpoint_model_dir
  elif FLAGS.restore_checkpoints:
    if not FLAGS.save_checkpoints:
      raise ValueError(
          'If restore_checkpoints is enabled, then save_checkpoints must be'
          ' enabled as well.'
      )
    checkpoint_model_dir = FLAGS.model_dir
  if checkpoint_model_dir is not None:
    logging.info('Restoring checkpoint from: %s', checkpoint_model_dir)
    training_state = training_state.restore_checkpoint(
        model_dir=checkpoint_model_dir)
    # Grab last step from the first of the optimizer replicas.
    start_step = int(training_state.optimizer.state.step[0])

  transformer_kwargs = training_state.transformer_kwargs
  metrics_all = []
  state_dict_summary_all = []
  state_dict_keys = get_state_dict_keys_from_flags()

  wrote_beam_hlo = False

  # Save the HLO and estimate compute / memory costs
  if jax.host_id() == 0:
    logging.info('Writing training HLO and estimating compute/memory costs.')
    state = training_state.flax_state
    optimizer = training_state.optimizer
    optimizer = optimizer.unreplicate()
    state = jax_utils.unreplicate(state)
    logging.info('Unreplicated optimizer and state.')
    _write_train_hlo(
        params=optimizer.target,
        state=state,
        hparams=hparams,
        transformer_kwargs=transformer_kwargs)
    logging.info('Wrote training HLO.')
    if FLAGS.estimate_compute_and_memory_cost:
      estimate_compute_and_memory_cost(
          FLAGS.model_dir,
          params=optimizer.target,
          state=state,
          hparams=hparams,
          transformer_kwargs=transformer_kwargs)
    state = jax_utils.replicate(state)
    optimizer = optimizer.replicate()
    logging.info('Replicated optimizer.')

  t_loop_start = time.time()
  for step, batch in zip(range(start_step, num_train_steps), train_iter):
    # Shard data to devices and do a training step.
    training_state, metrics = run_train_step(
        training_state=training_state, batch=batch, step=step, hparams=hparams)
    state = training_state.flax_state
    optimizer = training_state.optimizer
    metrics_all.append(metrics)
    state_dict_summary = summary_utils.get_state_dict_summary(
        state, state_dict_keys)
    state_dict_summary_all.append(state_dict_summary)

    # Save a Checkpoint
    if step % FLAGS.checkpoint_freq == 0 and step > 0:
      if jax.host_id() == 0 and FLAGS.save_checkpoints:
        training_state.save_checkpoint(model_dir=FLAGS.model_dir, step=step)


    # Periodic metric handling.
    if step % eval_freq == 0:

      # Training Metrics
      if FLAGS.compute_train_metrics:
        metrics_all = common_utils.get_metrics(metrics_all)
        state_dict_summary_all = common_utils.get_metrics(
            state_dict_summary_all)
        lr = metrics_all.pop('learning_rate').mean()

        metrics_sums = jax.tree.map(jnp.sum, metrics_all)
        denominator = metrics_sums.pop('denominator')
        summary = jax.tree.map(lambda x: x / denominator, metrics_sums)  # pylint: disable=cell-var-from-loop
        summary['learning_rate'] = lr
        if FLAGS.log_sparsity_scalars:
          apply_sparsity = metrics_all.pop('apply_sparsity').mean()
          update_weight_sparsity = metrics_all.pop(
              'update_weight_sparsity').mean()
          update_act_sparsity = metrics_all.pop('update_act_sparsity').mean()
          num_update_sparsity = metrics_all.pop('num_update_sparsity').mean()
          summary['apply_sparsity'] = apply_sparsity
          summary['update_weight_sparsity'] = update_weight_sparsity
          summary['update_act_sparsity'] = update_act_sparsity
          summary['num_update_sparsity'] = num_update_sparsity
        summary['perplexity'] = jnp.clip(jnp.exp(summary['loss']), a_max=1.0e4)
        logging.info('train in step: %d, loss: %.4f', step, summary['loss'])
        steps_per_eval = eval_freq if step != 0 else 1
        steps_per_sec = steps_per_eval / (time.time() - t_loop_start)
        t_loop_start = time.time()

        if jax.host_id() == 0:
          assert train_summary_writer is not None, ('train_summary_writer was '
                                                    'not initialized on host 0')
          train_summary_writer.scalar('steps per second', steps_per_sec, step)
          for key, val in summary.items():
            train_summary_writer.scalar(key, val, step)

          summary_utils.write_state_dict_summaries_to_tb(
              state_dict_summary_all, train_summary_writer,
              FLAGS.state_dict_summary_freq, step)
          train_summary_writer.flush()

      state_dict_summary_all = []
      metrics_all = []

      # TODO(shivaniagrawal): Add bleu score summaries for additional eval
      # datasets?
      # Additional eval metrics
      for eval_dataset_name, eval_ds in datasets.eval_ds_dict.items():
        summary_writer = None
        t_add_eval_start = time.time()
        if eval_dataset_name == FLAGS.eval_dataset_name:
          summary_writer = eval_summary_writer
        elif eval_dataset_name in additional_eval_summary_writers:
          summary_writer = additional_eval_summary_writers[eval_dataset_name]
        eval_summary = run_eval(
            ds=eval_ds,
            state=state,
            num_steps=num_eval_steps,
            params=optimizer.target,
            summary_writer=summary_writer,
            step=step,
            hparams=hparams,
            transformer_kwargs=transformer_kwargs)
        logging.info('eval for %s in step: %d, loss: %.4f', eval_dataset_name,
                     step, eval_summary['loss'])
        logging.info('eval time for %s: %.4f s step %d', eval_dataset_name,
                     time.time() - t_add_eval_start, step)

        # Save a checkpoint corresponding to when loss on the eval set is
        # minimized.
        if jax.host_id() == 0 and (eval_dataset_name == FLAGS.eval_dataset_name
                                   and FLAGS.save_minimum_loss_checkpoint):
          save_best_checkpoint(
              model_dir=FLAGS.model_dir,
              training_state=training_state,
              loss=eval_summary['loss'])

      if FLAGS.run_train_eval:
        # Eval-Train Metrics (Eval on the train set)
        # TODO(b/156127248): Might want to add odd-sized remainders to this
        # evaluation as well.
        t_train_eval_start = time.time()
        train_eval_summary = run_eval(
            ds=datasets.train_eval_ds,
            state=state,
            num_steps=num_eval_steps,
            params=optimizer.target,
            summary_writer=train_eval_summary_writer,
            step=step,
            hparams=hparams,
            transformer_kwargs=transformer_kwargs)
        logging.info('train eval in step: %d, loss: %.4f', step,
                     train_eval_summary['loss'])
        logging.info('train eval time: %.4f s step %d',
                     time.time() - t_train_eval_start, step)

      # Translation and BLEU Score.
      t_inference_start = time.time()
      should_write_beam_hlo = jax.host_id() == 0 and (not wrote_beam_hlo)
      sources, references, predictions = run_inference(
          ds=datasets.predict_ds_dict[FLAGS.eval_dataset_name],
          hparams=hparams,
          state=state,
          encoder=datasets.encoder,
          params=optimizer.target,
          write_beam_hlo=should_write_beam_hlo,
          transformer_kwargs=transformer_kwargs,
          step=step)
      if should_write_beam_hlo:
        wrote_beam_hlo = True
      logging.info('inference time: %.4f s step %d.',
                   time.time() - t_inference_start, step)
      logging.info('Translation: %d predictions %d references %d sources.',
                   len(predictions), len(references), len(sources))

      # Calculate BLEU score for translated eval corpus against reference.
      t_bleu_start = time.time()
      bleu_score = bleu.bleu_local(references, predictions)
      logging.info('bleu time: %.4f s step %d',
                   time.time() - t_bleu_start, step)
      t_bleu_start = time.time()
      sacrebleu_score = sacrebleu.corpus_bleu(predictions, [references]).score
      logging.info('sacrebleu time: %.4f s step %d',
                   time.time() - t_bleu_start, step)
      # Save translation samples for tensorboard.
      exemplars = ''
      for n in np.random.choice(np.arange(len(predictions)), 8):
        exemplars += f'{sources[n]}\n\n{references[n]}\n\n{predictions[n]}\n\n'
      if jax.host_id() == 0:
        assert eval_summary_writer is not None, ('eval_summary_writer was not '
                                                 'initialized on host 0')
        eval_summary_writer.scalar('bleu', bleu_score, step)
        eval_summary_writer.scalar('sacrebleu', sacrebleu_score, step)
        eval_summary_writer.text('samples', exemplars, step)
        eval_summary_writer.flush()



if __name__ == '__main__':
  app.run(main)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pandas_utils.py."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.utils import pandas_utils
import pandas as pd


class PandasUtilsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.df1 = pd.DataFrame({
        'name': ['aa', 'ab', 'cc', 'dd'],
        'run': [1, 1, 2, 3],
        'user': ['sam', 'max', 'max', 'jo']
    })

    self.df2 = pd.DataFrame({
        'name': ['aa', 'ab', 'cc'],
        'run': [1, 1, 2],
        'test1': ['x', 'y', 'z'],
        'test2': [10, 20, 30]
    })

    self.df3 = pd.DataFrame({
        'name': ['a', 'a', 'a', 'b', 'b', 'b'],
        'len': [2, 4, 6, 9, 10, 11],
        'dis': [1, 1, 1, 5, 6, 7],
    })

  @parameterized.named_parameters(
      dict(
          testcase_name='run_is_1',
          column_name='run',
          values=1,
          exp=pd.DataFrame(
              {
                  'name': ['aa', 'ab'],
                  'run': [1, 1],
                  'user': ['sam', 'max']
              },
              index=[0, 1])),
      dict(
          testcase_name='run_is_1_or_2',
          column_name='run',
          values=[1, 2],
          exp=pd.DataFrame(
              {
                  'name': ['aa', 'ab', 'cc'],
                  'run': [1, 1, 2],
                  'user': ['sam', 'max', 'max']
              },
              index=[0, 1, 2])),
  )
  def test_select_rows_by_column_values(self, column_name, values, exp):
    res = pandas_utils.select_rows_by_column_values(
        self.df1, column_name=column_name, values=values)
    pd.testing.assert_frame_equal(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='name_starts_with_a',
          column_name='name',
          regex_str='a.*',
          exp=pd.DataFrame({
              'name': ['aa', 'ab'],
              'run': [1, 1],
              'user': ['sam', 'max']
          })),
      dict(
          testcase_name='user_is_sam',
          column_name='user',
          regex_str='sam',
          exp=pd.DataFrame({
              'name': ['aa'],
              'run': [1],
              'user': ['sam']
          })),
  )
  def test_select_rows_by_regex(self, column_name, regex_str, exp):
    res = pandas_utils.select_rows_by_regex(
        self.df1, column_name=column_name, regex_str=regex_str)
    pd.testing.assert_frame_equal(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='run_is_not_1',
          column_name='run',
          values=[1],
          exp=pd.DataFrame(
              {
                  'name': ['cc', 'dd'],
                  'run': [2, 3],
                  'user': ['max', 'jo']
              },
              index=[2, 3])),
      dict(
          testcase_name='run_is_not_1_or_2',
          column_name='run',
          values=[1, 2],
          exp=pd.DataFrame({
              'name': ['dd'],
              'run': [3],
              'user': ['jo']
          },
                           index=[3])),
  )
  def test_drop_rows_by_column_values(self, column_name, values, exp):
    res = pandas_utils.drop_rows_by_column_values(
        self.df1, column_name=column_name, values=values)
    pd.testing.assert_frame_equal(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='name_does_not_start_with_a',
          column_name='name',
          regex_str='a.*',
          exp=pd.DataFrame(
              {
                  'name': ['cc', 'dd'],
                  'run': [2, 3],
                  'user': ['max', 'jo']
              },
              index=[2, 3])),
      dict(
          testcase_name='user_is_not_sam',
          column_name='user',
          regex_str='sam',
          exp=pd.DataFrame(
              {
                  'name': ['ab', 'cc', 'dd'],
                  'run': [1, 2, 3],
                  'user': ['max', 'max', 'jo']
              },
              index=[1, 2, 3])),
  )
  def test_drop_rows_by_regex(self, column_name, regex_str, exp):
    res = pandas_utils.drop_rows_by_regex(
        self.df1, column_name=column_name, regex_str=regex_str)
    pd.testing.assert_frame_equal(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='keep_name_column',
          columns_to_keep=['name'],
          exp=pd.DataFrame({
              'name': ['aa', 'ab', 'cc'],
          })),
      dict(
          testcase_name='keep_name_and_test2_columns',
          columns_to_keep=['run', 'test2'],
          exp=pd.DataFrame({
              'run': [1, 1, 2],
              'test2': [10, 20, 30]
          })),
  )
  def test_filter_columns(self, columns_to_keep, exp):
    res = pandas_utils.filter_columns(self.df2, columns_to_keep=columns_to_keep)
    pd.testing.assert_frame_equal(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='keep_columns_with_test_in_name',
          column_regex='test.*',
          exp=pd.DataFrame({
              'test1': ['x', 'y', 'z'],
              'test2': [10, 20, 30]
          })),
      dict(
          testcase_name='keep_columns_with_run_in_name',
          column_regex='run',
          exp=pd.DataFrame({
              'run': [1, 1, 2],
          })),
  )
  def test_filter_columns_by_regex(self, column_regex, exp):
    res = pandas_utils.filter_columns_by_regex(
        self.df2, column_regex=column_regex)
    pd.testing.assert_frame_equal(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='drop_columns_with_test_in_name',
          columns_to_drop=['run'],
          exp=pd.DataFrame({
              'name': ['aa', 'ab', 'cc'],
              'test1': ['x', 'y', 'z'],
              'test2': [10, 20, 30]
          })),
      dict(
          testcase_name='drop_columns_with_run_in_name',
          columns_to_drop=['run', 'test2'],
          exp=pd.DataFrame({
              'name': ['aa', 'ab', 'cc'],
              'test1': ['x', 'y', 'z'],
          })),
  )
  def test_drop_columns(self, columns_to_drop, exp):
    res = pandas_utils.drop_columns(self.df2, columns_to_drop=columns_to_drop)
    pd.testing.assert_frame_equal(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='drop_columns_with_test_in_name',
          column_regex='test.*',
          exp=pd.DataFrame({
              'name': ['aa', 'ab', 'cc'],
              'run': [1, 1, 2],
          })),
      dict(
          testcase_name='drop_columns_with_run_in_name',
          column_regex='run',
          exp=pd.DataFrame({
              'name': ['aa', 'ab', 'cc'],
              'test1': ['x', 'y', 'z'],
              'test2': [10, 20, 30]
          })),
  )
  def test_drop_columns_by_regex(self, column_regex, exp):
    res = pandas_utils.drop_columns_by_regex(
        self.df2, column_regex=column_regex)
    pd.testing.assert_frame_equal(res, exp)

  def test_group_by_with_aggregation(self):
    res = pandas_utils.group_by_with_aggregation(
        self.df3, by='name', agg_column_names=['len', 'dis'])
    self.assertEqual(res.shape, (2, 5))
    self.assertSequenceEqual(list(res.len['mean']), [4, 10])
    self.assertSequenceEqual(list(res.len['std']), [2, 1])
    self.assertSequenceEqual(list(res.dis['mean']), [1, 6])
    self.assertSequenceEqual(list(res.dis['std']), [0, 1])

  @parameterized.named_parameters(
      dict(
          testcase_name='simple_string',
          column_name='name',
          pattern='a',
          repl='d',
          exp=pd.DataFrame({
              'name': ['dd', 'db', 'cc', 'dd'],
              'run': [1, 1, 2, 3],
              'user': ['sam', 'max', 'max', 'jo']
          })),
      dict(
          testcase_name='regex_with_groups',
          column_name='user',
          pattern=r'(sa)',
          repl=r'\g<1>ttt',
          exp=pd.DataFrame({
              'name': ['aa', 'ab', 'cc', 'dd'],
              'run': [1, 1, 2, 3],
              'user': ['satttm', 'max', 'max', 'jo']
          })),
  )
  def test_rename_values_in_column(self, column_name, pattern, repl, exp):
    res = pandas_utils.rename_values_in_column(
        self.df1, column_name=column_name, pattern=pattern, repl=repl)
    pd.testing.assert_frame_equal(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='simple_string',
          pattern='user',
          repl='owner',
          exp=pd.DataFrame({
              'name': ['aa', 'ab', 'cc', 'dd'],
              'run': [1, 1, 2, 3],
              'owner': ['sam', 'max', 'max', 'jo']
          })),
      dict(
          testcase_name='regex_with_groups',
          pattern=r'(a)m',
          repl=r'x\g<1>',
          exp=pd.DataFrame({
              'nxae': ['aa', 'ab', 'cc', 'dd'],
              'run': [1, 1, 2, 3],
              'user': ['sam', 'max', 'max', 'jo']
          })),
  )
  def test_rename_column_headers(self, pattern, repl, exp):
    res = pandas_utils.rename_column_headers(
        self.df1, pattern=pattern, repl=repl)
    pd.testing.assert_frame_equal(res, exp)

  def test_apply_filter_drop_rename_operations(self):
    res = pandas_utils.apply_filter_drop_rename_operations(
        self.df1,
        row_filter_args=[('run', [1, 2])],
        row_regex_filter_args=[('user', r'ma.*')],
        rename_row_value_args=[('name', r'a(.*)', r'\g<1>')],
        drop_columns_by_regex_args=['user'],
        rename_column_name_args=[(r'(a)m', r'x\g<1>')],
        sort_by_args=([('run', False)]),
    )
    exp = pd.DataFrame({
        'nxae': ['cc', 'b'],
        'run': [2, 1],
    })
    pd.testing.assert_frame_equal(res, exp)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized dot_general."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

# pylint: disable=g-explicit-bool-comparison
# pylint: disable=g-explicit-length-test

import abc
import enum
from typing import Any, Sequence

from aqt.jax.v2 import aqt_quantizer
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import transpose
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import fp8_numerics
import jax
from jax import lax
from jax._src.numpy import lax_numpy
import jax.numpy as jnp
import numpy as onp
from typing_extensions import Self  # for python version < 3.11


dtypes_allowed_for_int32_accum = [jnp.int4, jnp.int8]


class CalibrationMode(enum.Enum):
  """Calibration axis modes."""

  CONTRACTING_AXIS = 1
  REMAINING_AXIS = 2


class DequantMode(enum.Enum):
  """Dequant modes."""

  # Multiply output of dot_general by the transposed scale
  # Compatible with CONTRACTING_AXIS.
  OUTPUT = 1
  # Multiply QTensor.qvalue by untransposed QTensor.scale before
  # dot_general (a.k.a. FakeQuant )
  # Compatible with both CalibrationMode.
  THIS_INPUT = 2
  # Multiply other argument of dot general by appropriately transposed scale.
  # Compatible with REMAINING_AXIS.
  OTHER_INPUT = 3


@utils.flax_slots_kw_only_dataclass
class Tensor:
  """Configuration of quantization of one tensor or one side of tensor op."""

  # Controls at what value of input tensor should be used.
  # Setting it to True, but not quantizing fwd pass will assert-fail.
  use_fwd_quant: bool = utils.static_field(default=False)
  # Dequantization mode.
  dequant_mode: DequantMode = utils.static_field(default=DequantMode.OUTPUT)
  # Calibration axis mode.
  calibration_mode: CalibrationMode = utils.static_field(
      default=CalibrationMode.CONTRACTING_AXIS
  )


@utils.flax_slots_kw_only_dataclass
class LocalAqt:
  contraction_axis_shard_count: None | int = utils.static_field(default=None)
  contraction_axis_shard_size: None | int = utils.static_field(default=None)
  # tile_largest_shape=True will apply the factor_reshape_largest function,
  # where shard_size is required and shard_count should be None.
  # If it is False, the original factor_reshape function will be applied,
  # where shard_count is required and shard_size should be None.
  tile_largest_shape: bool = utils.static_field(default=False)


def dot_general_raw_make(
    lhs_bits=None,
    rhs_bits=None,
    local_aqt=None,
    jax_scope_name='aqt',
    initialize_calibration=True,
    allow_dummy_gradients=False,
) -> 'DotGeneralRaw':
  """Create quantization configs for input matrices to a matmul."""
  # TODO: b/343490088 - Move all the parameters to dataclass defaults,
  #   provide setters to modify the configuration.
  lhs_cfg = Tensor()
  rhs_cfg = Tensor()

  # TODO(lew): Binary uses 0.5 right now, it should use -1 and 1.
  if (
      lhs_bits is not None
      and rhs_bits is not None
      and lhs_bits not in fp8_numerics.fp8_map.keys()
      and rhs_bits not in fp8_numerics.fp8_map.keys()
      and isinstance(lhs_bits, int)
      and 2 <= lhs_bits <= 8
      and isinstance(rhs_bits, int)
      and 2 <= rhs_bits <= 8
  ):
    dg_accumulator_dtype = jnp.int32
  else:
    dg_accumulator_dtype = None

  # DotGeneralRaw should create a this quantizer on default.
  # Then setter can change it.

  # initialize_calibration=False because that has to be delayed to be called
  # *inside* of flax.nn.custom_vjp
  lhs = aqt_quantizer.quantizer_make(
      lhs_bits, initialize_calibration=initialize_calibration
  )
  rhs = aqt_quantizer.quantizer_make(
      rhs_bits, initialize_calibration=initialize_calibration
  )
  # TODO(lew): This code (and surrounding code)is duplicated.
  # We should dedup.
  lhs_mid = aqt_quantizer.quantizer_make(
      None, initialize_calibration=initialize_calibration
  )
  rhs_mid = aqt_quantizer.quantizer_make(
      None, initialize_calibration=initialize_calibration
  )
  dg_quantizer = DefaultDotGeneralQuantizer(
      lhs=lhs, rhs=rhs, lhs_mid=lhs_mid, rhs_mid=rhs_mid
  )

  return DotGeneralRaw(
      lhs=lhs_cfg,
      rhs=rhs_cfg,
      dg_quantizer=dg_quantizer,
      dg_accumulator_dtype=dg_accumulator_dtype,
      local_aqt=local_aqt,
      jax_scope_name=jax_scope_name,
      allow_dummy_gradient_into_qtensor=allow_dummy_gradients,
  )


# TODO: b/343490088 - Move all the parameters to dataclass defaults,
#   provide setters to modify the configuration.
def dot_general_make(
    lhs_bits: None | int | fp8_numerics.FP8Dtype = None,
    rhs_bits: None | int | fp8_numerics.FP8Dtype = None,
    bwd_bits: None | int | fp8_numerics.FP8Dtype = None,
    use_fwd_quant: bool = True,
    dlhs_local_aqt=None,
    drhs_local_aqt=None,
    allow_dummy_gradients: bool = False,
) -> 'DotGeneral':
  """Create quantization configs for input matrices to a matmul."""
  fwd = dot_general_raw_make(
      lhs_bits,
      rhs_bits,
      jax_scope_name='aqt_fwd',
      initialize_calibration=False,
      allow_dummy_gradients=allow_dummy_gradients,
  )
  dlhs = dot_general_raw_make(
      bwd_bits,
      bwd_bits,
      local_aqt=dlhs_local_aqt,
      jax_scope_name='aqt_dlhs',
      initialize_calibration=False,
      allow_dummy_gradients=allow_dummy_gradients,
  )
  drhs = dot_general_raw_make(
      bwd_bits,
      bwd_bits,
      local_aqt=drhs_local_aqt,
      jax_scope_name='aqt_drhs',
      initialize_calibration=False,
      allow_dummy_gradients=allow_dummy_gradients,
  )
  cfg = DotGeneral(fwd=fwd, dlhs=dlhs, drhs=drhs)

  # Surprising: lhs quantization determines what drhs can do.
  if lhs_bits is not None:
    # Only rhs is accepting MultiTensor.
    cfg.drhs.rhs.use_fwd_quant = use_fwd_quant
  if rhs_bits is not None:
    cfg.dlhs.rhs.use_fwd_quant = use_fwd_quant
  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'
  return cfg


# ------------------------------------------------------------------------------


@utils.flax_slots_kw_only_dataclass
class MultiTensor:
  x: jnp.ndarray
  qx: aqt_tensor.QTensor


@utils.flax_slots_kw_only_dataclass
class TensorRes:
  """All the things we pass from the forward pass to the backward pass."""

  mt: MultiTensor
  quant_grad: aqt_tensor.GradientFn


@utils.flax_slots_kw_only_dataclass
class DotGeneralRes:
  lhs: TensorRes
  rhs: TensorRes


def einsum(eqn: str, lhs: jnp.ndarray, rhs: jnp.ndarray, dg=lax.dot_general):
  """A copy of jnp.einsum but without the default jit so as to be injectable."""
  operands, contractions = lax_numpy._default_poly_einsum_handler(  # pylint: disable=protected-access
      eqn, lhs, rhs, einsum_call=True, use_blas=True, optimize='optimal'
  )
  contractions = tuple((a, frozenset(b), c) for a, b, c, *_ in contractions)
  return jax.named_call(lax_numpy._einsum, name=eqn)(  # pylint: disable=protected-access
      operands,
      contractions,
      precision=None,
      preferred_element_type=None,
      _dot_general=dg,
  )


# TODO(lew): Inline and simplify, perhaps using list comprehension.
def _get_scale_t(
    qt: aqt_tensor.QTensor,
    transpose_fn: Any,
    dimension_numbers: lax.DotDimensionNumbers,
    lhs_shape: Sequence[int],
    rhs_shape: Sequence[int],
) -> Sequence[jnp.ndarray]:
  list_scale_t = []
  for scale in qt.scale:
    scale_t = transpose_fn(scale, dimension_numbers, lhs_shape, rhs_shape)
    list_scale_t.append(scale_t)
  return list_scale_t


def _apply_local_aqt(
    local_aqt: LocalAqt,
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    dimension_numbers: jax.lax.DotDimensionNumbers,
) -> tuple[jnp.ndarray, jnp.ndarray, jax.lax.DotDimensionNumbers]:
  """Applies local AQT if the configuration is given.

  Args:
    local_aqt: Local AQT configuration.
    lhs: Left hand side of the dot_general.
    rhs: Right hand side of the dot_general.
    dimension_numbers: Dot_general dimension numbers.

  Returns:
    A tuple of (lhs, rhs, dimension_numbers) with local AQT applied.
  """

  def factor_reshape_largest(x, ca, ba):
    # Tile the contraction axis with largest shape using tile_size.
    # This function can be a replacement for the original factor_reshape.
    # In the spirit of backward compatibility, we keep the original
    # factor_reshape as well.
    msg = 'When tile_largest_shape is True, must set shard_size.'
    assert local_aqt.contraction_axis_shard_size is not None, msg
    assert local_aqt.contraction_axis_shard_count is None, msg
    shard_size = local_aqt.contraction_axis_shard_size
    if not ca:
      return x, ca, ba
    shape = list(x.shape)
    max_ca_shape = max(shape[axis] for axis in ca)
    ax = shape.index(max_ca_shape)
    ax_ca_idx = ca.index(ax)
    assert max_ca_shape % shard_size == 0
    shape[ax] = max_ca_shape // shard_size
    shape.insert(ax + 1, shard_size)
    new_ca = [(b + int(b >= ax)) for b in ca]
    assert new_ca[ax_ca_idx] == ax + 1
    new_ba = [ax] + [(b + int(b > ax)) for b in ba]
    return x.reshape(shape), new_ca, new_ba

  def factor_reshape(x, ca, ba):
    # Original tiling function. Only support configuring the tile count.
    msg = 'You are using the original local aqt. Please only set shard_count.'
    assert local_aqt.contraction_axis_shard_count is not None, msg
    assert local_aqt.contraction_axis_shard_size is None, msg
    factor = local_aqt.contraction_axis_shard_count
    if not ca:
      return x, ca, ba
    shape = list(x.shape)
    ax = ca[0]
    orig_size = shape[ax]
    assert orig_size % factor == 0
    shape[ax] = factor
    shape.insert(ax + 1, orig_size // factor)
    new_ca = [(b + int(b >= ax)) for b in ca]
    assert new_ca[0] == ax + 1
    new_ba = [ax] + [(b + int(b > ax)) for b in ba]
    return x.reshape(shape), new_ca, new_ba

  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  if local_aqt.tile_largest_shape:
    factor_fn = factor_reshape_largest
  else:
    factor_fn = factor_reshape
  lhs, lhs_ca, lhs_ba = factor_fn(lhs, lhs_ca, lhs_ba)
  rhs, rhs_ca, rhs_ba = factor_fn(rhs, rhs_ca, rhs_ba)

  dimension_numbers = (lhs_ca, rhs_ca), (lhs_ba, rhs_ba)
  return lhs, rhs, dimension_numbers


@utils.flax_slots_kw_only_dataclass
class DotGeneralQuantizer(abc.ABC):
  """Abstract class for dot_general quantizer."""

  def __call__(
      self,
      lhs: jax.Array,
      rhs: jax.Array,
      dimension_numbers: jax.lax.DotDimensionNumbers,
      lhs_mode: CalibrationMode,
      rhs_mode: CalibrationMode,
  ) -> tuple[
      tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn],
      tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn],
  ]:
    (lhs, lhs_qt), (rhs, rhs_qt) = self.calibrate(
        lhs, rhs, dimension_numbers, lhs_mode, rhs_mode
    )
    return self.calculate_qvalue(lhs, lhs_qt, rhs, rhs_qt)

  @abc.abstractmethod
  def init_calibration(self):
    pass

  @abc.abstractmethod
  def calibrate(
      self,
      lhs: jax.Array,
      rhs: jax.Array,
      dimension_numbers: None | jax.lax.DotDimensionNumbers,
      lhs_mode: None | CalibrationMode,
      rhs_mode: None | CalibrationMode,
  ) -> tuple[
      tuple[jax.Array, aqt_tensor.QTensor], tuple[jax.Array, aqt_tensor.QTensor]
  ]:
    """Calculates incomplete QTensor from the given inputs.

    The function calculates incomplete QTensor from the given inputs. It also
    returns the updated lhs and rhs during calibration.

    Args:
      lhs: lhs
      rhs: rhs
      dimension_numbers: dimension_numbers
      lhs_mode: lhs_mode
      rhs_mode: rhs_mode

    Returns:
      A tuple of (lhs, lhs_qt), (rhs, rhs_qt) where lhs and rhs are updated
      arguments and lhs_qt and rhs_qt are incomplete QTensor.
    """
    pass

  # TODO(lew): There is only one meaningful implementation of
  # calculate_qvalue. Does not have to be an overridable method.
  @abc.abstractmethod
  def calculate_qvalue(
      self,
      lhs: jax.Array,
      lhs_qt: aqt_tensor.QTensor,
      rhs: jax.Array,
      rhs_qt: aqt_tensor.QTensor,
  ) -> tuple[
      tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn],
      tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn],
  ]:
    """Calculates qvalues from the given inputs."""
    pass

  @abc.abstractmethod
  def swap_lhs_and_rhs(self) -> None:
    """Swaps lhs and rhs configuration."""
    pass

  @abc.abstractmethod
  def assert_calib_shared_axes_value(
      self,
      lhs_val: None | Sequence[utils.AxisIdx],
      rhs_val: None | Sequence[utils.AxisIdx],
      msg: str,
  ) -> None:
    """Asserts if calib_shared_axes have certain values."""
    pass

  @abc.abstractmethod
  def set_context(
      self,
      lhs_context: utils.Context,
      rhs_context: utils.Context,
  ) -> None:
    """Sets context for lhs and rhs."""
    pass


# TODO(lew): Find a better name instead of Default* (Factored?, Separated?)
@utils.flax_slots_kw_only_dataclass
class DefaultDotGeneralQuantizer(DotGeneralQuantizer):
  """Default dot_general quantizer."""

  lhs: aqt_quantizer.Quantizer
  rhs: aqt_quantizer.Quantizer

  # Quantizers for "middle" scale in "matmul" like SmoothQuant, AWQ, etc.
  lhs_mid: aqt_quantizer.Quantizer
  rhs_mid: aqt_quantizer.Quantizer

  # The amount (exponent) of the scales that should be transferred to the
  # other side. 0.0 = nothing, 1.0 = all.
  lhs_mid_alpha: None | float = None
  rhs_mid_alpha: None | float = None

  # This is a hack to make QTensors compatible with the current
  # _qtensor_dot_general.
  # The QTensors that are returned do not include the mid-scales.
  # But this is ok, because the skipped mid-scales are reciprocal of each other
  # and they would cancel out in _qtensor_dot_general anyway.
  # A good design would be to hardcode skip_mid_scales=False because it would
  # maintain semantics of QTensor (QTensor.dequant).
  # This also is needed to send a correct QTensor to backprop
  # in  use_fwd_quant=True mode.
  # We don't do it now mostly because it would require a separate mechanism
  # in _qtensor_dot_general to skip the mid-scales
  # (which do cancel each other mathematically).
  skip_mid_scales: bool = True

  def init_calibration(self):
    self.lhs.init_calibration()
    self.rhs.init_calibration()
    if self.lhs_mid_alpha is not None:
      self.lhs_mid.init_calibration()
    if self.rhs_mid_alpha is not None:
      self.rhs_mid.init_calibration()

  def calibrate(
      self,
      lhs: jax.Array,
      rhs: jax.Array,
      dimension_numbers: None | jax.lax.DotDimensionNumbers,
      lhs_mode: None | CalibrationMode,
      rhs_mode: None | CalibrationMode,
  ) -> tuple[
      tuple[jax.Array, aqt_tensor.QTensor], tuple[jax.Array, aqt_tensor.QTensor]
  ]:
    if dimension_numbers is not None:
      (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
      lhs_ra = utils.get_remaining_axes(lhs.ndim, lhs_ca, lhs_ba)
      rhs_ra = utils.get_remaining_axes(rhs.ndim, rhs_ca, rhs_ba)

      def _get_calibration_axes(
          mode: CalibrationMode,
          ndim: int,
          ca: Sequence[utils.AxisIdx],
          ba: Sequence[utils.AxisIdx],
      ) -> Sequence[utils.AxisIdx]:
        """Computes calibration axes for the given Tensor."""
        match mode:
          case CalibrationMode.REMAINING_AXIS:
            calibration_axes = utils.get_remaining_axes(ndim, ca, ba)
            m = 'mid-quantization is not supported for REMAINING_AXIS mode'
            assert self.lhs_mid_alpha is None and self.rhs_mid_alpha is None, m
          case CalibrationMode.CONTRACTING_AXIS:
            calibration_axes = ca
          case _:
            raise ValueError(f'Unknown calibration mode: {mode}')
        return calibration_axes

      lhs_calib_axes = _get_calibration_axes(lhs_mode, lhs.ndim, lhs_ca, lhs_ba)
      rhs_calib_axes = _get_calibration_axes(rhs_mode, rhs.ndim, rhs_ca, rhs_ba)
    else:
      (lhs_ra, rhs_ra) = (None, None)
      lhs_calib_axes = None
      rhs_calib_axes = None

    def dezero(x):
      return jnp.where(x == 0.0, jnp.ones_like(x), x)

    if self.lhs_mid_alpha is not None:
      assert self.lhs_mid is not None
      lhs_mid_qt = self.lhs_mid.calibrate(lhs, calibration_axes=lhs_ra)
      assert len(lhs_mid_qt.scale) == 1, 'you must set some numerics'
      lhs_mid_scale = dezero(lhs_mid_qt.scale[0])
      lhs_mid_scale = lhs_mid_scale**self.lhs_mid_alpha
      lhs_mid_scale_t = transpose.lhs_scale_transpose_for_rhs_input(
          lhs_mid_scale, dimension_numbers, rhs.shape
      )
    else:
      lhs_mid_scale = 1.0
      lhs_mid_scale_t = 1.0

    if self.rhs_mid_alpha is not None:
      assert self.rhs_mid is not None
      rhs_mid_qt = self.rhs_mid.calibrate(rhs, calibration_axes=rhs_ra)
      assert len(rhs_mid_qt.scale) == 1, 'you must set some numerics'
      rhs_mid_scale = dezero(rhs_mid_qt.scale[0])
      rhs_mid_scale = rhs_mid_scale**self.rhs_mid_alpha
      rhs_mid_scale_t = transpose.rhs_scale_transpose_for_lhs_input(
          rhs_mid_scale, dimension_numbers, lhs.shape
      )
    else:
      rhs_mid_scale = 1.0
      rhs_mid_scale_t = 1.0

    # This condition can be considered an optimization.
    # ATM it also allows us to not deal with 1.0 being a scalar.
    if self.lhs_mid_alpha is not None or self.rhs_mid_alpha is not None:
      # Combined SmoothQuant scales
      lhs_mid_scale_combined = lhs_mid_scale / rhs_mid_scale_t
      rhs_mid_scale_combined = rhs_mid_scale / lhs_mid_scale_t

      # Apply the combined scales before per-tensor calibration
      lhs_mid = lhs / lhs_mid_scale_combined
      rhs_mid = rhs / rhs_mid_scale_combined

      # Per-tensor calibration, same as "else" branch.
      lhs_qt = self.lhs.calibrate(lhs_mid, calibration_axes=lhs_calib_axes)
      rhs_qt = self.rhs.calibrate(rhs_mid, calibration_axes=rhs_calib_axes)

      # To maintain QTensor.dequant semantics, we need to append the combined
      # scales.
      assert lhs_qt.scale is not None
      assert rhs_qt.scale is not None
      if not self.skip_mid_scales:
        lhs_qt.scale.append(lhs_mid_scale_combined)
        rhs_qt.scale.append(rhs_mid_scale_combined)
    else:
      lhs_qt = self.lhs.calibrate(lhs, calibration_axes=lhs_calib_axes)
      rhs_qt = self.rhs.calibrate(rhs, calibration_axes=rhs_calib_axes)

    return ((lhs, lhs_qt), (rhs, rhs_qt))

  def calculate_qvalue(
      self,
      lhs: jax.Array,
      lhs_qt: aqt_tensor.QTensor,
      rhs: jax.Array,
      rhs_qt: aqt_tensor.QTensor,
  ) -> tuple[
      tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn],
      tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn],
  ]:
    """Calculates qvalues from the given inputs."""
    lhs_qt, lhs_grad = self.lhs.calculate_qvalue(lhs, lhs_qt)
    rhs_qt, rhs_grad = self.rhs.calculate_qvalue(rhs, rhs_qt)

    return (lhs_qt, lhs_grad), (rhs_qt, rhs_grad)

  def swap_lhs_and_rhs(self) -> None:
    self.lhs, self.rhs = self.rhs, self.lhs

  def assert_calib_shared_axes_value(
      self,
      lhs_val: None | Sequence[utils.AxisIdx],
      rhs_val: None | Sequence[utils.AxisIdx],
      msg: str,
  ) -> None:
    assert self.lhs.calib_shared_axes == lhs_val, msg
    assert self.rhs.calib_shared_axes == rhs_val, msg

  def set_context(
      self,
      lhs_context: utils.Context,
      rhs_context: utils.Context,
  ) -> None:
    self.lhs.context = lhs_context
    self.rhs.context = rhs_context


def quant(
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    lhs_qt: None | aqt_tensor.QTensor,
    rhs_qt: None | aqt_tensor.QTensor,
    dg_quantizer: DotGeneralQuantizer,
    lhs_cfg: Tensor,
    rhs_cfg: Tensor,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    allow_dummy_gradient_into_qtensor: bool,
) -> tuple[
    tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn],
    tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn],
]:
  """Quantizes the given lhs and rhs using dg_quantizer."""

  def _postprocess_qtensor(
      input_qtensor: None | aqt_tensor.QTensor,
      calculated_qtensor: aqt_tensor.QTensor,
      quant_grad: aqt_tensor.GradientFn,
  ) -> tuple[aqt_tensor.QTensor, str | aqt_tensor.GradientFn]:
    """Compute qtensor from input or input_qtensor."""
    # TODO(lew): moving this if out of DotGeneralRaw into v2.flax.DotGeneral
    # would induce huge code simplification here.
    # E.g. this file would not have to know about DotGeneralQuantizer.
    if input_qtensor is not None:
      if not allow_dummy_gradient_into_qtensor:
        quant_grad = (
            'Poison. '
            + 'Gradients are not generally expected in serving. '
            + 'Please set allow_dummy_gradient_into_qtensor to True '
            + 'if this is the intended behavior.'
        )
      output_qtensor = input_qtensor
    else:
      output_qtensor = calculated_qtensor

    return output_qtensor, quant_grad

  (lhs, lhs_incomplete_qt), (rhs, rhs_incomplete_qt) = dg_quantizer.calibrate(
      lhs,
      rhs,
      dimension_numbers,
      lhs_cfg.calibration_mode,
      rhs_cfg.calibration_mode,
  )

  if lhs_qt is not None and not lhs_qt.is_full():
    # Incomplete QTensor is provided as lhs_qt.
    lhs_incomplete_qt = lhs_qt
    lhs_qt = None

  if rhs_qt is not None and not rhs_qt.is_full():
    # Incomplete QTensor is provided as rhs_qt.
    rhs_incomplete_qt = rhs_qt
    rhs_qt = None

  lhs_quantized, rhs_quantized = dg_quantizer.calculate_qvalue(
      lhs, lhs_incomplete_qt, rhs, rhs_incomplete_qt
  )
  lhs_qt_calculated, lhs_quant_grad = lhs_quantized
  rhs_qt_calculated, rhs_quant_grad = rhs_quantized

  lhs_qt, lhs_quant_grad = _postprocess_qtensor(
      lhs_qt,
      lhs_qt_calculated,
      lhs_quant_grad,
  )

  rhs_qt, rhs_quant_grad = _postprocess_qtensor(
      rhs_qt,
      rhs_qt_calculated,
      rhs_quant_grad,
  )
  return (lhs_qt, lhs_quant_grad), (rhs_qt, rhs_quant_grad)


def _maybe_use_fwd_quant(
    lhs: jnp.ndarray,
    rhs: MultiTensor,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    use_fwd_quant: bool,
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Applies already quantized value for backpropagation, if the flag is set.

  Note that we cannot directly use the quantized rhs for gradient calculation,
  since the dimension numbers are changed.

  Args:
    lhs: Left hand size of gradient dot_general.
    rhs: Right hand size of gradient dot_general.
    dimension_numbers: Dot_general dimension numbers.
    use_fwd_quant: Flag to use forward quantization.

  Returns:
    A tuple of updated (lhs, rhs). If use_fwd_quant is True, lhs is multiplied
    with rhs scale, while rhs is set to the original rhs's qvalue.
  """
  scale_count = -1
  if rhs.qx.scale is not None:
    scale_count = len(rhs.qx.scale)
  msg = (
      f'Found use_fwd_quant is {use_fwd_quant} in bwd. '
      'It is supported only if there is exactly one scale in a good shape.\n'
      f'{scale_count=}'
  )
  if use_fwd_quant:
    assert scale_count == 1, msg
    if rhs.qx.bias:
      raise NotImplementedError(
          'Quantization biases are not supported in forward quantization.'
      )

    # It this transpose fails or the multpilication below fails,
    # we have some misconfiguration. One way to deal with it is
    # set use_fwd_quant to False.
    scale_t = transpose.rhs_scale_transpose_for_lhs_input(
        rhs.qx.scale[0], dimension_numbers, lhs.shape
    )

    # Cast rhs scales to lhs dtype when multiplying with lhs. This is to
    # avoid an unexpected upcast when rhs is float32 but lhs is float16.
    lhs = lhs * scale_t.astype(lhs.dtype)

    # rhs qvalue may be integer. It will be quantized again later, so cast
    # its dtype back to dequant dtype.
    # TODO(yichizh): avoid double quantization and evaluate model quality.
    rhs = rhs.qx.qvalue.astype(rhs.qx.dequant_dtype)
  else:
    rhs = rhs.x

  return lhs, rhs


@utils.flax_slots_kw_only_dataclass
class DotGeneralRaw:
  """Configuration of quantization of one dot_general without gradient."""

  lhs: Tensor
  rhs: Tensor
  dg_quantizer: DotGeneralQuantizer
  dg_accumulator_dtype: None | jnp.dtype = utils.static_field()
  local_aqt: None | LocalAqt = utils.static_field()
  jax_scope_name: str = utils.static_field()

  # Set it to true in order to train with non-None lhs_qt or rhs_qt.
  # Gradient will still flow into lhs_qt and/or rhs_qt, but it may be incorrect.
  # It is a caller responsibility to NOT update these QTensors
  allow_dummy_gradient_into_qtensor: bool = utils.static_field(default=False)
  dot_general: utils.DotGeneralT = utils.static_field(
      default=jax.lax.dot_general
  )

  # TODO(lew): Remove this function.
  @classmethod
  def make(cls, *args, **kwargs) -> Self:
    return dot_general_raw_make(*args, **kwargs)

  # TODO(lew): Can we remove MutliTensor and pass rhs_qt instead?
  def __call__(
      self,
      lhs: jnp.ndarray,
      rhs: jnp.ndarray | MultiTensor,
      # xhs_qt are used in serving.
      lhs_qt: None | aqt_tensor.QTensor,
      rhs_qt: None | aqt_tensor.QTensor,
      dimension_numbers: jax.lax.DotDimensionNumbers,
  ):
    """A quantized dot_general function without custom gradient."""
    with jax.named_scope(self.jax_scope_name):
      # TODO(lew):
      #  - Use qx.value with the int type.
      #  - Handle qx.value with the int type in an optimized way.
      #  - Add a "FQ" case we multiply qx.value*qx.value_scale (not transposed).
      #  - Can we carry untransposed scale and transpose here?
      if isinstance(rhs, MultiTensor):
        lhs, rhs = _maybe_use_fwd_quant(
            lhs, rhs, dimension_numbers, self.rhs.use_fwd_quant
        )
      assert isinstance(rhs, jnp.ndarray)

      # TODO(lew): Define cutsom_vjp on tiled_dot_general and replace local_aqt.
      if self.local_aqt is not None:
        msg = 'Custom calib_shared_axes not implemented for local AQT.'
        if isinstance(self.dg_quantizer, DefaultDotGeneralQuantizer):
          self.dg_quantizer.assert_calib_shared_axes_value(None, None, msg)

        lhs, rhs, dimension_numbers = _apply_local_aqt(
            self.local_aqt,  # pytype: disable=attribute-error
            lhs,
            rhs,
            dimension_numbers,
        )

      (lhs_qt, lhs_quant_grad), (rhs_qt, rhs_quant_grad) = quant(
          lhs,
          rhs,
          lhs_qt,
          rhs_qt,
          self.dg_quantizer,
          self.lhs,
          self.rhs,
          dimension_numbers,
          self.allow_dummy_gradient_into_qtensor,
      )

      msg = (
          'biases are only supported in fake quant mode, but got a {arg} bias '
          'and self.{arg}.dequant_mode == {mode} != DequantMode.THIS_INPUT'
      )
      assert not (
          lhs_qt.bias and self.lhs.dequant_mode != DequantMode.THIS_INPUT
      ), msg.format(arg='lhs', mode=self.lhs.dequant_mode)
      assert not (
          rhs_qt.bias and self.rhs.dequant_mode != DequantMode.THIS_INPUT
      ), msg.format(arg='rhs', mode=self.rhs.dequant_mode)

      lhs_mt = MultiTensor(x=lhs, qx=lhs_qt)
      lhs_res = TensorRes(mt=lhs_mt, quant_grad=lhs_quant_grad)

      rhs_mt = MultiTensor(x=rhs, qx=rhs_qt)
      rhs_res = TensorRes(mt=rhs_mt, quant_grad=rhs_quant_grad)

      # TODO(lew): mt.x above should be clipped for clipping calibrations
      out = _qtensor_dot_general(
          lhs_qt, rhs_qt, dimension_numbers, self, jnp.promote_types(lhs, rhs)
      )

      out = out.dequant()

      res = DotGeneralRes(lhs=lhs_res, rhs=rhs_res)
      if self.local_aqt is not None:
        (lhs_ca, rhs_ca), _ = dimension_numbers
        assert len(lhs_ca) == len(rhs_ca)
        if len(lhs_ca) > 0:
          out = jnp.sum(out, axis=0)
        # We are not supporting local AQT in fwd pass, so no res needed.
        res = None
      return out, res


def _qtensor_dot_general(
    lhs_qt: aqt_tensor.QTensor,
    rhs_qt: aqt_tensor.QTensor,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    cfg: Any,
    # dequant_dtype: DType,
    dequant_dtype: jnp.dtype,
) -> aqt_tensor.QTensor:
  """QTensor lax.dot_general replacement."""

  def _maybe_dequant(
      input_qtensor: aqt_tensor.QTensor, tensor_cfg: Tensor
  ) -> jnp.ndarray:
    if tensor_cfg.dequant_mode == DequantMode.THIS_INPUT:
      output = input_qtensor.dequant()
    else:
      output = input_qtensor.qvalue
    return output

  # Dequantize before the lax dg call if in fake quant mode
  lhs_qin = _maybe_dequant(lhs_qt, cfg.lhs)
  rhs_qin = _maybe_dequant(rhs_qt, cfg.rhs)

  dtype_ms = (
      f'Found {cfg.dg_accumulator_dtype=}, {lhs_qin.dtype=} and'
      f' {rhs_qin.dtype=}. Dot general accumulator dtype can only be'
      ' jnp.int32 when both inputs are int8. Otherwise it is recommended to'
      ' be None to let lax.dot_general automatically decide it.'
  )
  if cfg.dg_accumulator_dtype == jnp.int32:
    assert (
        lhs_qin.dtype in dtypes_allowed_for_int32_accum
        and rhs_qin.dtype in dtypes_allowed_for_int32_accum
    ), dtype_ms

  dtypes_can_be_scaled = [jnp.bfloat16, jnp.float32, jnp.float64]

  # If the transposes below fail it might be because of a misconfiguration.
  # For instance "mid" quantization in DefaultDotGeneralQuantizer is not
  # compatible with DequantMode.OTHER_INPUT
  # TODO(lew): One way to deal with it is to have a per-scale DequantMode.

  if cfg.lhs.dequant_mode == DequantMode.OTHER_INPUT:
    assert rhs_qin.dtype in dtypes_can_be_scaled
    for scale in lhs_qt.scale:
      transposed_scale = transpose.lhs_scale_transpose_for_rhs_input(
          scale, dimension_numbers, rhs_qt.shape
      )
      assert isinstance(transposed_scale, jnp.ndarray)  # make pytype quiet
      rhs_qin = rhs_qin * transposed_scale.astype(rhs_qin.dtype)

  if cfg.rhs.dequant_mode == DequantMode.OTHER_INPUT:
    assert lhs_qin.dtype in dtypes_can_be_scaled
    for scale in rhs_qt.scale:
      transposed_scale = transpose.rhs_scale_transpose_for_lhs_input(
          scale, dimension_numbers, lhs_qt.shape
      )
      assert isinstance(transposed_scale, jnp.ndarray)  # make pytype quiet
      lhs_qin = lhs_qin * transposed_scale.astype(lhs_qin.dtype)

  if jax.local_devices()[0].platform == 'cpu':
    # needed bet lax.dot_general(int4, int4) is illegal on cpu.
    # TODO(aqt): Remove this platform check once
    # https://github.com/google/jax/issues/19682 is fixed.
    # TODO(yichizh): It's better to assert False here with the following msg
    # msg = (
    #     'lax.dot_general(int4, int4) is illegal on cpu:'
    #     ' https://github.com/google/jax/issues/19682. The simple workaround'
    #     ' is to upcast to int8, but in that case please directly set the'
    #     ' numerics bits to int8. Please contact the AQT team if you believe'
    #     ' the workaround is needed.'
    # )
    if lhs_qin.dtype == jnp.int4 and rhs_qin.dtype == jnp.int4:
      lhs_qin = lhs_qin.astype(jnp.int8)
      rhs_qin = rhs_qin.astype(jnp.int8)

  out = cfg.dot_general(
      lhs_qin,
      rhs_qin,
      dimension_numbers=dimension_numbers,
      preferred_element_type=cfg.dg_accumulator_dtype,
      precision=lax.Precision.DEFAULT,
  )
  # TODO(lew): Do we have a correct precision above?
  #   Relevant: https://github.com/google/jax/issues/14022
  out = aqt_tensor.QTensor(
      qvalue=out,
      scale=[],
      scale_t=None,
      bias=[],
      dequant_dtype=dequant_dtype,
  )
  assert out.scale is not None  # pytype help

  if cfg.lhs.dequant_mode == DequantMode.OUTPUT:
    extend_scale = _get_scale_t(
        qt=lhs_qt,
        transpose_fn=transpose.lhs_scale_transpose_to_output,
        dimension_numbers=dimension_numbers,
        lhs_shape=lhs_qin.shape,
        rhs_shape=rhs_qin.shape,
    )

    out.scale.extend(extend_scale)
  if cfg.rhs.dequant_mode == DequantMode.OUTPUT:
    extend_scale = _get_scale_t(
        qt=rhs_qt,
        transpose_fn=transpose.rhs_scale_transpose_to_output,
        dimension_numbers=dimension_numbers,
        lhs_shape=lhs_qin.shape,
        rhs_shape=rhs_qin.shape,
    )
    out.scale.extend(extend_scale)
  return out


@utils.flax_slots_kw_only_dataclass
class DotGeneral:
  """Configuration of quantization of dot_general and its gradients."""

  fwd: DotGeneralRaw
  dlhs: DotGeneralRaw
  drhs: DotGeneralRaw

  apply_custom_vjp_on_jax: bool = utils.static_field(default=True)

  @classmethod
  def make(cls, *args, **kwargs) -> Self:
    return dot_general_make(*args, **kwargs)

  def dg_core(
      self,
      lhs: jnp.ndarray,
      rhs: jnp.ndarray,
      lhs_qt: None | aqt_tensor.QTensor,
      rhs_qt: None | aqt_tensor.QTensor,
      dimension_numbers: lax.DotDimensionNumbers,
  ):
    """dot_general function with expanded API."""
    msg = 'AQT is not yet optimized to accept quantized types directly. '
    msg += f'lhs.dtype: {lhs.dtype}, rhs.dtype: {rhs.dtype}'
    assert lhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg
    assert rhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg

    self.assert_config_validity()
    # TODO(dhchoi): Refactor this to make both branches to have the similar
    # functionality of applying custom_vjp.
    if self.apply_custom_vjp_on_jax:
      dg_core_with_custom_grad = jax.custom_vjp(_dg_core, nondiff_argnums=(4,))
      dg_core_with_custom_grad.defvjp(dg_core_vjp_fwd, dg_core_vjp_bwd)
      out, res = dg_core_with_custom_grad(
          lhs, rhs, lhs_qt, rhs_qt, dimension_numbers, self
      )
    else:
      out, res = _dg_core(lhs, rhs, lhs_qt, rhs_qt, dimension_numbers, self)  # pytype: disable=wrong-arg-types
    return out, res

  def __call__(
      self,
      lhs,
      rhs,
      dimension_numbers,
      precision=None,
      preferred_element_type=None,
  ):
    del preferred_element_type
    assert (
        precision is None
    ), f'Precision {precision} requested together with quantization.'

    out, _ = self.dg_core(
        lhs=lhs,
        rhs=rhs,
        lhs_qt=None,
        rhs_qt=None,
        dimension_numbers=dimension_numbers,
    )
    return out

  def assert_config_validity(self: Self):
    """Asserts if configuration is valid."""

    # use_fwd_quant is not enabled when calibration_axis = remaining_axis.
    # TODO: b/336198483 - Enable use_fwd_quant flag in the case.
    expected_fwd_quant = False
    msg_fwd_quant = (
        f'use_fwd_quant should be set to {expected_fwd_quant} when remaining'
        ' axis are used for calibration axis. '
    )

    if self.fwd.rhs.calibration_mode == CalibrationMode.REMAINING_AXIS:
      msg_fwd_quant += (
          f'rhs.calibration_mode: {self.fwd.rhs.calibration_mode},'
          f' dlhs use_fwd_quant: {self.dlhs.rhs.use_fwd_quant}'
      )
      assert self.dlhs.rhs.use_fwd_quant is expected_fwd_quant, msg_fwd_quant

    if self.fwd.lhs.calibration_mode == CalibrationMode.REMAINING_AXIS:
      msg_fwd_quant += (
          f'lhs.calibration_mode: {self.fwd.lhs.calibration_mode},'
          f' drhs use_fwd_quant: {self.drhs.rhs.use_fwd_quant}'
      )
      assert self.drhs.rhs.use_fwd_quant is expected_fwd_quant, msg_fwd_quant

    # Check valid combination between calibration_mode and dequant_mode
    unsupported_calibration_dequant_pairs = [
        (DequantMode.OUTPUT, CalibrationMode.REMAINING_AXIS),
        (DequantMode.OTHER_INPUT, CalibrationMode.CONTRACTING_AXIS),
    ]
    msg_mode_mismatch = (
        'Unsupported calibration mode - dequant mode combination '
    )
    for (
        dequant_mode,
        calibration_mode,
    ) in unsupported_calibration_dequant_pairs:
      assert not (
          self.fwd.lhs.calibration_mode == calibration_mode
          and self.fwd.lhs.dequant_mode == dequant_mode
      ), (
          msg_mode_mismatch
          + ' for lhs. calibration_mode:'
          f' {self.fwd.lhs.calibration_mode}, dequant_mode:'
          f' {self.fwd.lhs.dequant_mode}'
      )
      assert not (
          self.fwd.rhs.calibration_mode == calibration_mode
          and self.fwd.rhs.dequant_mode == dequant_mode
      ), (
          msg_mode_mismatch
          + ' for rhs. calibration_mode:'
          f' {self.fwd.rhs.calibration_mode}, dequant_mode:'
          f' {self.fwd.rhs.dequant_mode}'
      )


def _dg_core(
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    lhs_qt: None | aqt_tensor.QTensor,
    rhs_qt: None | aqt_tensor.QTensor,
    dimension_numbers: lax.DotDimensionNumbers,
    cfg: DotGeneral,
):
  out, _ = dg_core_vjp_fwd(lhs, rhs, lhs_qt, rhs_qt, dimension_numbers, cfg)
  return out


# When defining a vjp, all traceable variables must be input arguments of
# both the fwd and bwd function.
# The cfg (DotGeneral) contains the key used for stochastic rounding,
# which are traceable dynamic variables. It needs to be an input argument
# to prevent the jax side effect.
def dg_core_vjp_fwd(
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    lhs_qt: None | aqt_tensor.QTensor,
    rhs_qt: None | aqt_tensor.QTensor,
    dimension_numbers: lax.DotDimensionNumbers,
    cfg: DotGeneral,
):
  """custom_vjp fwd pass."""
  assert (
      lhs.dtype == rhs.dtype
  ), f'Unmatched lhs and rhs dtype: {lhs.dtype} vs {rhs.dtype}'
  cfg.fwd.dg_quantizer.init_calibration()
  cfg.dlhs.dg_quantizer.init_calibration()
  cfg.drhs.dg_quantizer.init_calibration()
  ret, res = cfg.fwd(
      lhs,
      rhs,
      lhs_qt,
      rhs_qt,
      dimension_numbers,
  )
  ret = ret.astype(lhs.dtype)
  # We return these values to allow for materialization.
  assert res is not None, 'res cannot be None in fwd pass.'
  qret = (res.lhs.mt.qx, res.rhs.mt.qx)
  return ((ret, qret), (res, cfg))


def _update_dimension_numbers_for_backward(
    fwd_dimension_numbers: jax.lax.DotDimensionNumbers,
    y_is_lhs: bool,
    gradient_rank: int,
    y_rank: int,
) -> tuple[jax.lax.DotDimensionNumbers, tuple[int, ...]]:
  """Generates a new dimension number for backward pass.

  For dot_general(lhs, rhs), the gradients for lhs and rhs are calculated by
  dot_general(g, rhs) and dot_general(g, lhs). This function generates proper
  dimension numbers for the dot_generals used for gradient calculation.

  Args:
    fwd_dimension_numbers: Dimension number used during forward pass
    y_is_lhs: If set, the function calculates dimension numbers for dlhs.
    gradient_rank: Rank of the gradient.
    y_rank: Rank of the other side input.

  Returns:
    A tuple of (dimension numbers for gradient dot_general, transpose axes to be
    applied on the gradient dot_generals output to match with the original
    argument dimension).
  """

  def ranges_like(*xs):
    start = 0
    for x in xs:
      yield tuple(range(start, start + len(x)))
      start += len(x)

  (x_ca, y_ca), (x_ba, y_ba) = fwd_dimension_numbers
  if y_is_lhs:
    (y_ca, x_ca) = (x_ca, y_ca)
    (y_ba, x_ba) = (x_ba, y_ba)

  gradient_rank = gradient_rank - y_rank + len(x_ba) + 2 * len(x_ca)
  x_ra = tuple(utils.get_remaining_axes(gradient_rank, x_ca, x_ba))
  y_ra = tuple(utils.get_remaining_axes(y_rank, y_ca, y_ba))
  if y_is_lhs:
    g_ba, g_ca, _ = ranges_like(x_ba, y_ra, x_ra)
  else:
    g_ba, _, g_ca = ranges_like(x_ba, x_ra, y_ra)
  dims = ((g_ca, y_ra), (g_ba, y_ba))

  x_ca_sorted_by_y = tuple(onp.take(x_ca, onp.argsort(y_ca)))
  out_transpose_axes = tuple(onp.argsort(tuple(x_ba) + x_ra + x_ca_sorted_by_y))

  return dims, out_transpose_axes


def dg_core_vjp_bwd(
    fwd_dimension_numbers: lax.DotDimensionNumbers,
    res: tuple[None | DotGeneralRes, DotGeneral],
    g,
):
  """custom_vjp bwd pass."""
  dg_res, cfg = res
  msg = 'dg_res can only be None in 2nd derivative. It is not yet supported.'
  assert dg_res is not None, msg
  g = g[0]  # g[1] is gradient with respect to qret which we are ignoring.

  def grad_dot_general(
      y_res: TensorRes,
      quant_grad: aqt_tensor.GradientFn,
      dg_raw: DotGeneralRaw,
      y_is_lhs,
  ):
    dims, out_transpose_axes = _update_dimension_numbers_for_backward(
        fwd_dimension_numbers, y_is_lhs, g.ndim, y_res.mt.x.ndim
    )

    out, _ = dg_raw(g, y_res.mt, None, None, dims)

    transposed_out = jax.lax.transpose(out, out_transpose_axes)
    if quant_grad is not None:
      transposed_out = quant_grad(transposed_out)[0]
    return transposed_out

  dlhs = grad_dot_general(
      dg_res.rhs,
      dg_res.lhs.quant_grad,
      cfg.dlhs,
      False,
  )
  drhs = grad_dot_general(
      dg_res.lhs,
      dg_res.rhs.quant_grad,
      cfg.drhs,
      True,
  )
  # fwd_dimension_numbers are marked as nondiff_argnums instead of returning
  # None as grad to it. This is because it is a tuple of Python integers
  # that cannot be traced by Jax.
  return (dlhs, drhs, None, None, None)
